@@ -1,14 +1,10 @@
 import os
 import sys
 import json
-import time
-import pytest
 import uuid
-import httpx
 import asyncio
 
 from pymixin.mixin_bot_api import MixinBotApi
-from pymixin import mixin_config
 from pymixin import log
 
 logger = log.get_logger(__name__)
@@ -27,7 +23,9 @@ async def start():
     hint = str(uuid.uuid4())
     r = await mixin_bot.verify_pin()
     print(r)
+    await mixin_bot.close()
     # r = await mixin_bot.get_ghost_keys(['f32bb6ab-1a4d-44bb-892f-f0129239df89'], 0, hint)
     # print(r)
 
-asyncio.run(start())
\ No newline at end of file
+asyncio.run(start())
+
