@@ -240,7 +240,7 @@ func buildRawTransactionCmd(c *cli.Context) error {
 		return err
 	}
 
-	inputs := make([]map[string]interface{}, 0)
+	inputs := make([]map[string]any, 0)
 	for _, in := range strings.Split(c.String("inputs"), ",") {
 		parts := strings.Split(in, ":")
 		if len(parts) != 2 {
@@ -254,13 +254,13 @@ func buildRawTransactionCmd(c *cli.Context) error {
 		if err != nil {
 			return err
 		}
-		inputs = append(inputs, map[string]interface{}{
+		inputs = append(inputs, map[string]any{
 			"hash":  hash,
 			"index": int(index),
 		})
 	}
 
-	outputs := make([]map[string]interface{}, 0)
+	outputs := make([]map[string]any, 0)
 	for _, out := range strings.Split(c.String("outputs"), ",") {
 		parts := strings.Split(out, ":")
 		if len(parts) != 2 {
@@ -274,7 +274,7 @@ func buildRawTransactionCmd(c *cli.Context) error {
 		if amount.Sign() == 0 {
 			return fmt.Errorf("invalid output %s", out)
 		}
-		outputs = append(outputs, map[string]interface{}{
+		outputs = append(outputs, map[string]any{
 			"accounts": []*common.Address{&addr},
 			"amount":   amount,
 		})
@@ -282,10 +282,10 @@ func buildRawTransactionCmd(c *cli.Context) error {
 
 	var raw signerInput
 	raw.Node = c.String("node")
-	isb, _ := json.Marshal(map[string]interface{}{"inputs": inputs})
+	isb, _ := json.Marshal(map[string]any{"inputs": inputs})
 	json.Unmarshal(isb, &raw)
 
-	tx := common.NewTransaction(asset)
+	tx := common.NewTransactionV3(asset)
 	for _, in := range inputs {
 		tx.AddInput(in["hash"].(crypto.Hash), in["index"].(int))
 	}
@@ -294,7 +294,7 @@ func buildRawTransactionCmd(c *cli.Context) error {
 	}
 	tx.Extra = extra
 
-	signed := tx.AsLatestVersion()
+	signed := tx.AsVersioned()
 	for i := range tx.Inputs {
 		err = signed.SignInput(raw, i, []*common.Address{&account})
 		if err != nil {
@@ -325,7 +325,7 @@ func signTransactionCmd(c *cli.Context) error {
 		}
 	}
 
-	tx := common.NewTransaction(raw.Asset)
+	tx := common.NewTransactionV3(raw.Asset)
 	for _, in := range raw.Inputs {
 		if d := in.Deposit; d != nil {
 			tx.AddDepositInput(&common.DepositData{
@@ -384,7 +384,7 @@ func signTransactionCmd(c *cli.Context) error {
 		accounts = append(accounts, &account)
 	}
 
-	signed := tx.AsLatestVersion()
+	signed := tx.AsVersioned()
 	for i := range signed.Inputs {
 		err := signed.SignInput(raw, i, accounts)
 		if err != nil {
@@ -396,7 +396,7 @@ func signTransactionCmd(c *cli.Context) error {
 }
 
 func sendTransactionCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "sendrawtransaction", []interface{}{
+	data, err := callRPC(c.String("node"), "sendrawtransaction", []any{
 		c.String("raw"),
 	}, c.Bool("time"))
 	if err == nil {
@@ -448,12 +448,12 @@ func pledgeNodeCmd(c *cli.Context) error {
 
 	amount := common.NewIntegerFromString(c.String("amount"))
 
-	tx := common.NewTransaction(common.XINAssetId)
+	tx := common.NewTransactionV3(common.XINAssetId)
 	tx.AddInput(input, 0)
 	tx.AddOutputWithType(common.OutputTypeNodePledge, nil, common.Script{}, amount, seed)
 	tx.Extra = append(signer.PublicSpendKey[:], payee.PublicSpendKey[:]...)
 
-	signed := tx.AsLatestVersion()
+	signed := tx.AsVersioned()
 	err = signed.SignInput(raw, 0, []*common.Address{&account})
 	if err != nil {
 		return err
@@ -525,7 +525,7 @@ func cancelNodeCmd(c *cli.Context) error {
 		return fmt.Errorf("invalid source and receiver %s %s", pig.String(), receiver.PublicSpendKey)
 	}
 
-	tx := common.NewTransaction(common.XINAssetId)
+	tx := common.NewTransactionV3(common.XINAssetId)
 	tx.AddInput(pledge.PayloadHash(), 0)
 	tx.AddOutputWithType(common.OutputTypeNodeCancel, nil, common.Script{}, pledge.Outputs[0].Amount.Div(100), seed)
 	tx.AddScriptOutput([]*common.Address{&receiver}, common.NewThresholdScript(1), pledge.Outputs[0].Amount.Sub(tx.Outputs[0].Amount), seed)
@@ -541,7 +541,7 @@ func cancelNodeCmd(c *cli.Context) error {
 			Mask: source.Outputs[0].Mask,
 		},
 	}
-	signed := tx.AsLatestVersion()
+	signed := tx.AsVersioned()
 	err = signed.SignUTXO(utxo, []*common.Address{&account})
 	if err != nil {
 		return err
@@ -584,7 +584,7 @@ func decodePledgeNodeCmd(c *cli.Context) error {
 }
 
 func getRoundLinkCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "getroundlink", []interface{}{
+	data, err := callRPC(c.String("node"), "getroundlink", []any{
 		c.String("from"),
 		c.String("to"),
 	}, c.Bool("time"))
@@ -595,7 +595,7 @@ func getRoundLinkCmd(c *cli.Context) error {
 }
 
 func getRoundByNumberCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "getroundbynumber", []interface{}{
+	data, err := callRPC(c.String("node"), "getroundbynumber", []any{
 		c.String("id"),
 		c.Uint64("number"),
 	}, c.Bool("time"))
@@ -606,7 +606,7 @@ func getRoundByNumberCmd(c *cli.Context) error {
 }
 
 func getRoundByHashCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "getroundbyhash", []interface{}{
+	data, err := callRPC(c.String("node"), "getroundbyhash", []any{
 		c.String("hash"),
 	}, c.Bool("time"))
 	if err == nil {
@@ -616,7 +616,7 @@ func getRoundByHashCmd(c *cli.Context) error {
 }
 
 func listSnapshotsCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "listsnapshots", []interface{}{
+	data, err := callRPC(c.String("node"), "listsnapshots", []any{
 		c.Uint64("since"),
 		c.Uint64("count"),
 		c.Bool("sig"),
@@ -629,7 +629,7 @@ func listSnapshotsCmd(c *cli.Context) error {
 }
 
 func getSnapshotCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "getsnapshot", []interface{}{
+	data, err := callRPC(c.String("node"), "getsnapshot", []any{
 		c.String("hash"),
 	}, c.Bool("time"))
 	if err == nil {
@@ -639,7 +639,7 @@ func getSnapshotCmd(c *cli.Context) error {
 }
 
 func getTransactionCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "gettransaction", []interface{}{
+	data, err := callRPC(c.String("node"), "gettransaction", []any{
 		c.String("hash"),
 	}, c.Bool("time"))
 	if err == nil {
@@ -649,7 +649,7 @@ func getTransactionCmd(c *cli.Context) error {
 }
 
 func getCacheTransactionCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "getcachetransaction", []interface{}{
+	data, err := callRPC(c.String("node"), "getcachetransaction", []any{
 		c.String("hash"),
 	}, c.Bool("time"))
 	if err == nil {
@@ -659,7 +659,7 @@ func getCacheTransactionCmd(c *cli.Context) error {
 }
 
 func getUTXOCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "getutxo", []interface{}{
+	data, err := callRPC(c.String("node"), "getutxo", []any{
 		c.String("hash"),
 		c.Uint64("index"),
 	}, c.Bool("time"))
@@ -669,8 +669,18 @@ func getUTXOCmd(c *cli.Context) error {
 	return err
 }
 
+func getKeyCmd(c *cli.Context) error {
+	data, err := callRPC(c.String("node"), "getkey", []any{
+		c.String("key"),
+	}, c.Bool("time"))
+	if err == nil {
+		fmt.Println(string(data))
+	}
+	return err
+}
+
 func listMintWorksCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "listmintworks", []interface{}{
+	data, err := callRPC(c.String("node"), "listmintworks", []any{
 		c.Uint64("since"),
 	}, c.Bool("time"))
 	if err == nil {
@@ -680,7 +690,7 @@ func listMintWorksCmd(c *cli.Context) error {
 }
 
 func listMintDistributionsCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "listmintdistributions", []interface{}{
+	data, err := callRPC(c.String("node"), "listmintdistributions", []any{
 		c.Uint64("since"),
 		c.Uint64("count"),
 		c.Bool("tx"),
@@ -692,7 +702,7 @@ func listMintDistributionsCmd(c *cli.Context) error {
 }
 
 func listAllNodesCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "listallnodes", []interface{}{
+	data, err := callRPC(c.String("node"), "listallnodes", []any{
 		c.Uint64("threshold"),
 		c.Bool("state"),
 	}, c.Bool("time"))
@@ -703,7 +713,7 @@ func listAllNodesCmd(c *cli.Context) error {
 }
 
 func getInfoCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "getinfo", []interface{}{}, c.Bool("time"))
+	data, err := callRPC(c.String("node"), "getinfo", []any{}, c.Bool("time"))
 	if err == nil {
 		fmt.Println(string(data))
 	}
@@ -711,7 +721,7 @@ func getInfoCmd(c *cli.Context) error {
 }
 
 func dumpGraphHeadCmd(c *cli.Context) error {
-	data, err := callRPC(c.String("node"), "dumpgraphhead", []interface{}{}, c.Bool("time"))
+	data, err := callRPC(c.String("node"), "dumpgraphhead", []any{}, c.Bool("time"))
 	if err == nil {
 		fmt.Println(string(data))
 	}
@@ -745,7 +755,7 @@ func setupTestNetCmd(c *cli.Context) error {
 			"balance": "10000",
 		})
 	}
-	genesis := map[string]interface{}{
+	genesis := map[string]any{
 		"epoch": time.Now().Unix(),
 		"nodes": inputs,
 		"domains": []map[string]string{
@@ -802,12 +812,12 @@ peers = [%s]
 
 var httpClient *http.Client
 
-func callRPC(node, method string, params []interface{}, pt bool) ([]byte, error) {
+func callRPC(node, method string, params []any, pt bool) ([]byte, error) {
 	if httpClient == nil {
 		httpClient = &http.Client{Timeout: 60 * time.Second}
 	}
 
-	body, err := json.Marshal(map[string]interface{}{
+	body, err := json.Marshal(map[string]any{
 		"method": method,
 		"params": params,
 	})
@@ -834,8 +844,8 @@ func callRPC(node, method string, params []interface{}, pt bool) ([]byte, error)
 
 	var result struct {
 		Runtime string      `json:"runtime"`
-		Data    interface{} `json:"data"`
-		Error   interface{} `json:"error"`
+		Data    any `json:"data"`
+		Error   any `json:"error"`
 	}
 	dec := json.NewDecoder(resp.Body)
 	dec.UseNumber()
@@ -892,7 +902,7 @@ func (raw signerInput) ReadUTXOKeys(hash crypto.Hash, index int) (*common.UTXOKe
 		}
 	}
 
-	data, err := callRPC(raw.Node, "getutxo", []interface{}{hash.String(), index}, false)
+	data, err := callRPC(raw.Node, "getutxo", []any{hash.String(), index}, false)
 	if err != nil {
 		return nil, err
 	}
@@ -917,32 +927,32 @@ func (raw signerInput) ReadLastMintDistribution(group string) (*common.MintDistr
 	return nil, nil
 }
 
-func transactionToMap(tx *common.VersionedTransaction) map[string]interface{} {
-	var inputs []map[string]interface{}
+func transactionToMap(tx *common.VersionedTransaction) map[string]any {
+	var inputs []map[string]any
 	for _, in := range tx.Inputs {
 		if in.Hash.HasValue() {
-			inputs = append(inputs, map[string]interface{}{
+			inputs = append(inputs, map[string]any{
 				"hash":  in.Hash,
 				"index": in.Index,
 			})
 		} else if len(in.Genesis) > 0 {
-			inputs = append(inputs, map[string]interface{}{
+			inputs = append(inputs, map[string]any{
 				"genesis": hex.EncodeToString(in.Genesis),
 			})
 		} else if in.Deposit != nil {
-			inputs = append(inputs, map[string]interface{}{
+			inputs = append(inputs, map[string]any{
 				"deposit": in.Deposit,
 			})
 		} else if in.Mint != nil {
-			inputs = append(inputs, map[string]interface{}{
+			inputs = append(inputs, map[string]any{
 				"mint": in.Mint,
 			})
 		}
 	}
 
-	var outputs []map[string]interface{}
+	var outputs []map[string]any
 	for _, out := range tx.Outputs {
-		output := map[string]interface{}{
+		output := map[string]any{
 			"type":   out.Type,
 			"amount": out.Amount,
 		}
@@ -956,7 +966,7 @@ func transactionToMap(tx *common.VersionedTransaction) map[string]interface{} {
 			output["mask"] = out.Mask
 		}
 		if w := out.Withdrawal; w != nil {
-			output["withdrawal"] = map[string]interface{}{
+			output["withdrawal"] = map[string]any{
 				"chain":     w.Chain,
 				"asset_key": w.AssetKey,
 				"address":   w.Address,
@@ -966,7 +976,7 @@ func transactionToMap(tx *common.VersionedTransaction) map[string]interface{} {
 		outputs = append(outputs, output)
 	}
 
-	tm := map[string]interface{}{
+	tm := map[string]any{
 		"version": tx.Version,
 		"asset":   tx.Asset,
 		"inputs":  inputs,
@@ -975,7 +985,7 @@ func transactionToMap(tx *common.VersionedTransaction) map[string]interface{} {
 		"hash":    tx.PayloadHash(),
 	}
 	if as := tx.AggregatedSignature; as != nil {
-		tm["aggregated"] = map[string]interface{}{
+		tm["aggregated"] = map[string]any{
 			"signers":   as.Signers,
 			"signature": as.Signature,
 		}
