@@ -6,6 +6,7 @@ import (
 	"fmt"
 	"strconv"
 
+	"github.com/zeebo/blake3"
 	"golang.org/x/crypto/sha3"
 )
 
@@ -15,6 +16,10 @@ func NewHash(data []byte) Hash {
 	return Hash(sha3.Sum256(data))
 }
 
+func Blake3Hash(data []byte) Hash {
+	return Hash(blake3.Sum256(data))
+}
+
 func HashFromString(src string) (Hash, error) {
 	var hash Hash
 	data, err := hex.DecodeString(src)
