@@ -9,7 +9,7 @@ import (
 )
 
 func (node *Node) Loop() error {
-	rand.Seed(time.Now().UnixNano())
+	rand.Seed(clock.Now().UnixNano())
 	err := node.PingNeighborsFromConfig()
 	if err != nil {
 		return err
@@ -38,7 +38,7 @@ func (node *Node) Teardown() {
 	node.chains.RUnlock()
 	node.Peer.Teardown()
 	node.persistStore.Close()
-	node.cacheStore.Reset()
+	node.cacheStore.Clear()
 }
 
 func TestMockReset() {
