@@ -19,6 +19,9 @@ const (
 	CosiActionExternalAnnouncement
 	CosiActionExternalChallenge
 	CosiActionFinalization
+	CosiActionExternalCommitments
+	CosiActionSelfFullCommitment
+	CosiActionExternalFullChallenge
 )
 
 type CosiAction struct {
@@ -31,6 +34,9 @@ type CosiAction struct {
 	Response     *[32]byte
 	Transaction  *common.VersionedTransaction
 	WantTx       bool
+	Commitments  []*crypto.Key
+	Challenge    *crypto.Key
+	random       *crypto.Key
 	finalized    bool
 	data         *CosiChainData
 }
@@ -43,11 +49,12 @@ type CosiChainData struct {
 }
 
 type CosiAggregator struct {
-	Snapshot    *common.Snapshot
-	Transaction *common.VersionedTransaction
-	WantTxs     map[crypto.Hash]bool
-	Commitments map[int]*crypto.Key
-	Responses   map[int]*[32]byte
+	Snapshot       *common.Snapshot
+	Transaction    *common.VersionedTransaction
+	WantTxs        map[crypto.Hash]bool
+	FullChallenges map[crypto.Hash]bool
+	Commitments    map[int]*crypto.Key
+	Responses      map[int]*[32]byte
 }
 
 type CosiVerifier struct {
@@ -56,6 +63,18 @@ type CosiVerifier struct {
 	random     *crypto.Key
 }
 
+func (node *Node) cosiAcceptedNodesList(ts uint64) []*CNode {
+	var nodes []*CNode
+	for _, n := range node.NodesListWithoutState(ts, true) {
+		nodes = append(nodes, n)
+	}
+	for i := range nodes {
+		j := int(ts % uint64(i+1))
+		nodes[i], nodes[j] = nodes[j], nodes[i]
+	}
+	return nodes
+}
+
 func (chain *Chain) cosiHook(m *CosiAction) (bool, error) {
 	logger.Debugf("cosiHook(%s) %v\n", chain.ChainId, m)
 	if !chain.running {
@@ -71,8 +90,8 @@ func (chain *Chain) cosiHook(m *CosiAction) (bool, error) {
 	if m.finalized || !m.WantTx || m.PeerId == chain.node.IdForNetwork {
 		return m.finalized, nil
 	}
-	logger.Debugf("cosiHook finalized snapshot without transaction %s %s %s\n", m.PeerId, m.SnapshotHash, m.Snapshot.Transaction)
-	chain.node.Peer.SendTransactionRequestMessage(m.PeerId, m.Snapshot.Transaction)
+	logger.Debugf("cosiHook finalized snapshot without transaction %s %s %s\n", m.PeerId, m.SnapshotHash, m.Snapshot.SoleTransaction())
+	chain.node.Peer.SendTransactionRequestMessage(m.PeerId, m.Snapshot.SoleTransaction())
 	return m.finalized, nil
 }
 
@@ -80,6 +99,9 @@ func (chain *Chain) cosiHandleAction(m *CosiAction) error {
 	if m.Action == CosiActionFinalization {
 		return chain.cosiHandleFinalization(m)
 	}
+	if m.Action == CosiActionExternalCommitments {
+		return chain.cosiAddCommitments(m)
+	}
 	if err := chain.checkActionSanity(m); err != nil {
 		logger.Debugf("cosiHandleAction checkActionSanity %v ERROR %s\n", m, err)
 		return nil
@@ -88,7 +110,7 @@ func (chain *Chain) cosiHandleAction(m *CosiAction) error {
 	switch m.Action {
 	case CosiActionSelfEmpty:
 		return chain.cosiSendAnnouncement(m)
-	case CosiActionSelfCommitment:
+	case CosiActionSelfCommitment, CosiActionSelfFullCommitment:
 		return chain.cosiHandleCommitment(m)
 	case CosiActionSelfResponse:
 		return chain.cosiHandleResponse(m)
@@ -96,6 +118,8 @@ func (chain *Chain) cosiHandleAction(m *CosiAction) error {
 		return chain.cosiHandleAnnouncement(m)
 	case CosiActionExternalChallenge:
 		return chain.cosiHandleChallenge(m)
+	case CosiActionExternalFullChallenge:
+		return chain.cosiHandleFullChallenge(m)
 	}
 
 	return nil
@@ -114,7 +138,7 @@ func (chain *Chain) checkActionSanity(m *CosiAction) error {
 		if s.Signature != nil || s.Timestamp != 0 {
 			return fmt.Errorf("only empty snapshot can be announced")
 		}
-	case CosiActionSelfCommitment, CosiActionSelfResponse:
+	case CosiActionSelfCommitment, CosiActionSelfFullCommitment, CosiActionSelfResponse:
 		if chain.ChainId != chain.node.IdForNetwork {
 			return fmt.Errorf("self action aggregation chain %s %s", chain.ChainId, chain.node.IdForNetwork)
 		}
@@ -134,9 +158,28 @@ func (chain *Chain) checkActionSanity(m *CosiAction) error {
 		if s.Signature != nil || s.Timestamp == 0 {
 			return fmt.Errorf("only empty snapshot with timestamp can be announced")
 		}
-		ov := chain.CosiVerifiers[s.Transaction]
+		ov := chain.CosiVerifiers[s.SoleTransaction()]
 		if ov != nil && s.RoundNumber > 0 && ov.Snapshot.RoundNumber == s.RoundNumber && s.Timestamp < ov.Snapshot.Timestamp+config.SnapshotRoundGap {
-			return fmt.Errorf("a transaction %s only in one round %d of one chain %s", s.Transaction, s.RoundNumber, chain.ChainId)
+			return fmt.Errorf("a transaction %s only in one round %d of one chain %s", s.SoleTransaction(), s.RoundNumber, chain.ChainId)
+		}
+	case CosiActionExternalFullChallenge:
+		if chain.ChainId == chain.node.IdForNetwork {
+			return fmt.Errorf("external action announcement chain %s %s", chain.ChainId, chain.node.IdForNetwork)
+		}
+		if chain.ChainId != m.PeerId {
+			return fmt.Errorf("external action announcement peer %s %s", chain.ChainId, m.PeerId)
+		}
+		if s.Signature != nil || s.Timestamp == 0 || m.Challenge == nil {
+			return fmt.Errorf("only empty snapshot with timestamp and challenge can be fully challenged")
+		}
+		m.random = chain.cosiRetrieveRandom(m.SnapshotHash, m.PeerId, m.Challenge)
+		if m.random == nil {
+			err := chain.cosiPrepareRandomsAndSendCommitments(m.PeerId)
+			return fmt.Errorf("no match random for the commitment %v %v", m, err)
+		}
+		ov := chain.CosiVerifiers[s.SoleTransaction()]
+		if ov != nil && s.RoundNumber > 0 && ov.Snapshot.RoundNumber == s.RoundNumber && s.Timestamp < ov.Snapshot.Timestamp+config.SnapshotRoundGap {
+			return fmt.Errorf("a transaction %s only in one round %d of one chain %s", s.SoleTransaction(), s.RoundNumber, chain.ChainId)
 		}
 	case CosiActionExternalChallenge:
 		if chain.ChainId == chain.node.IdForNetwork {
@@ -153,7 +196,7 @@ func (chain *Chain) checkActionSanity(m *CosiAction) error {
 	if s == nil {
 		return fmt.Errorf("no snapshot in cosi")
 	}
-	if s.Version != common.SnapshotVersion {
+	if s.Version != common.SnapshotVersionCommonEncoding {
 		return fmt.Errorf("invalid snapshot version %d", s.Version)
 	}
 	if s.NodeId != chain.ChainId {
@@ -202,6 +245,11 @@ func (chain *Chain) checkActionSanity(m *CosiAction) error {
 		return fmt.Errorf("node is slow in catching up")
 	}
 
+	rn := chain.node.GetRemovingOrSlashingNode(m.PeerId)
+	if rn != nil {
+		return fmt.Errorf("peer node %s is removing or slashing", m.PeerId)
+	}
+
 	cn := chain.node.GetAcceptedOrPledgingNode(chain.ChainId)
 	if cn == nil {
 		return fmt.Errorf("chain node %s not found", chain.ChainId)
@@ -250,7 +298,7 @@ func (chain *Chain) cosiSendAnnouncement(m *CosiAction) error {
 			if err != nil {
 				return err
 			}
-			best := chain.determinBestRound(s.Timestamp)
+			best := chain.determineBestRound(s.Timestamp)
 			threshold := external.Timestamp + config.SnapshotReferenceThreshold*config.SnapshotRoundGap*36
 			if best != nil && best.NodeId != final.NodeId && threshold < best.Start {
 				logger.Verbosef("CosiLoop cosiHandleAction cosiSendAnnouncement new best external %s:%d:%d => %s:%d:%d\n", external.NodeId, external.Number, external.Timestamp, best.NodeId, best.Number, best.Start)
@@ -263,7 +311,7 @@ func (chain *Chain) cosiSendAnnouncement(m *CosiAction) error {
 				return chain.clearAndQueueSnapshotOrPanic(s)
 			}
 		} else if start, _ := cache.Gap(); s.Timestamp >= start+config.SnapshotRoundGap {
-			best := chain.determinBestRound(s.Timestamp)
+			best := chain.determineBestRound(s.Timestamp)
 			if best == nil {
 				logger.Verbosef("CosiLoop cosiHandleAction cosiSendAnnouncement no best available\n")
 				return chain.clearAndQueueSnapshotOrPanic(s)
@@ -297,34 +345,52 @@ func (chain *Chain) cosiSendAnnouncement(m *CosiAction) error {
 		s.References = cache.References
 	}
 
-	ov := chain.CosiVerifiers[s.Transaction]
+	ov := chain.CosiVerifiers[s.SoleTransaction()]
 	if ov != nil && s.RoundNumber > 0 && ov.Snapshot.RoundNumber == s.RoundNumber && s.Timestamp < ov.Snapshot.Timestamp+config.SnapshotRoundGap {
-		err := fmt.Errorf("a transaction %s only in one round %d of one chain %s", s.Transaction, s.RoundNumber, chain.ChainId)
+		err := fmt.Errorf("a transaction %s only in one round %d of one chain %s", s.SoleTransaction(), s.RoundNumber, chain.ChainId)
 		logger.Verbosef("CosiLoop cosiHandleAction cosiSendAnnouncement ERROR %s\n", err)
 		return nil
 	}
 
 	s.Hash = s.PayloadHash()
 	agg := &CosiAggregator{
-		Snapshot:    s,
-		Transaction: cd.TX,
-		WantTxs:     make(map[crypto.Hash]bool),
-		Commitments: make(map[int]*crypto.Key),
-		Responses:   make(map[int]*[32]byte),
+		Snapshot:       s,
+		Transaction:    cd.TX,
+		WantTxs:        make(map[crypto.Hash]bool),
+		FullChallenges: make(map[crypto.Hash]bool),
+		Commitments:    make(map[int]*crypto.Key),
+		Responses:      make(map[int]*[32]byte),
 	}
+
 	v := &CosiVerifier{Snapshot: s, random: crypto.CosiCommit(rand.Reader)}
 	R := v.random.Public()
 	chain.CosiVerifiers[s.Hash] = v
-	chain.CosiVerifiers[s.Transaction] = v
+	chain.CosiVerifiers[s.SoleTransaction()] = v
 	agg.Commitments[cd.CN.ConsensusIndex] = &R
 	chain.CosiAggregators[s.Hash] = agg
-	nodes := chain.node.NodesListWithoutState(s.Timestamp, true)
+	nodes := chain.node.cosiAcceptedNodesList(s.Timestamp)
 	for _, cn := range nodes {
 		peerId := cn.IdForNetwork
-		err := chain.node.Peer.SendSnapshotAnnouncementMessage(peerId, m.Snapshot, R)
-		if err != nil {
-			logger.Verbosef("CosiLoop cosiHandleAction cosiSendAnnouncement SendSnapshotAnnouncementMessage(%s, %s) ERROR %s\n", peerId, s.Hash, err.Error())
+		if peerId == chain.ChainId {
+			continue
 		}
+		commitment := chain.cosiPopCommitment(peerId)
+		if commitment == nil {
+			err := chain.node.Peer.SendSnapshotAnnouncementMessage(peerId, m.Snapshot, R)
+			if err != nil {
+				logger.Verbosef("CosiLoop cosiHandleAction cosiSendAnnouncement SendSnapshotAnnouncementMessage(%s, %s) ERROR %v\n", peerId, s.Hash, err)
+			}
+			continue
+		}
+		cam := &CosiAction{
+			PeerId:       peerId,
+			Action:       CosiActionSelfFullCommitment,
+			SnapshotHash: s.Hash,
+			Commitment:   commitment,
+			Snapshot:     s,
+			WantTx:       true,
+		}
+		chain.AppendCosiAction(cam)
 	}
 	return nil
 }
@@ -381,10 +447,14 @@ func (chain *Chain) cosiHandleAnnouncement(m *CosiAction) error {
 	r := crypto.CosiCommit(rand.Reader)
 	v := &CosiVerifier{Snapshot: s, Commitment: m.Commitment, random: r}
 	chain.CosiVerifiers[s.Hash] = v
-	chain.CosiVerifiers[s.Transaction] = v
+	chain.CosiVerifiers[s.SoleTransaction()] = v
 	err := chain.node.Peer.SendSnapshotCommitmentMessage(s.NodeId, s.Hash, r.Public(), cd.TX == nil)
 	if err != nil {
-		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleAnnouncement SendSnapshotCommitmentMessage(%s, %s) ERROR %s\n", s.NodeId, s.Hash, err.Error())
+		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleAnnouncement SendSnapshotCommitmentMessage(%s, %s) ERROR %v\n", s.NodeId, s.Hash, err)
+	}
+	err = chain.cosiPrepareRandomsAndSendCommitments(s.NodeId)
+	if err != nil {
+		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleAnnouncement SendCommitmentsMessage(%s) ERROR %v\n", s.NodeId, err)
 	}
 	return nil
 }
@@ -398,13 +468,14 @@ func (chain *Chain) cosiHandleCommitment(m *CosiAction) error {
 		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleCommitment %v REPEAT\n", m)
 		return nil
 	}
-	base := chain.node.ConsensusThreshold(ann.Snapshot.Timestamp)
+	base := chain.node.ConsensusThreshold(ann.Snapshot.Timestamp, false)
 	if len(ann.Commitments) >= base {
 		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleCommitment %v EXCEED\n", m)
 		return nil
 	}
 	ann.Commitments[cd.PN.ConsensusIndex] = m.Commitment
 	ann.WantTxs[m.PeerId] = m.WantTx
+	ann.FullChallenges[m.PeerId] = m.Action == CosiActionSelfFullCommitment
 	logger.Verbosef("CosiLoop cosiHandleAction cosiHandleCommitment %v NOW %d %d\nn", m, len(ann.Commitments), base)
 	if len(ann.Commitments) < base {
 		return nil
@@ -426,10 +497,12 @@ func (chain *Chain) cosiHandleCommitment(m *CosiAction) error {
 	ann.Responses[cd.CN.ConsensusIndex] = response
 	copy(cosi.Signature[32:], response[:])
 
-	nodes := chain.node.NodesListWithoutState(s.Timestamp, true)
+	nodes := chain.node.cosiAcceptedNodesList(s.Timestamp)
 	for _, cn := range nodes {
 		id := cn.IdForNetwork
-		if wantTx, found := ann.WantTxs[id]; !found {
+		if ann.FullChallenges[id] {
+			err = chain.node.Peer.SendFullChallengeMessage(id, s, ann.Commitments[cd.CN.ConsensusIndex], ann.Commitments[cn.ConsensusIndex], cd.TX)
+		} else if wantTx, found := ann.WantTxs[id]; !found {
 			continue
 		} else if wantTx {
 			err = chain.node.Peer.SendTransactionChallengeMessage(id, m.SnapshotHash, cosi, cd.TX)
@@ -437,12 +510,79 @@ func (chain *Chain) cosiHandleCommitment(m *CosiAction) error {
 			err = chain.node.Peer.SendTransactionChallengeMessage(id, m.SnapshotHash, cosi, nil)
 		}
 		if err != nil {
-			logger.Verbosef("CosiLoop cosiHandleAction cosiHandleCommitment SendTransactionChallengeMessage(%s, %s) ERROR %s\n", id, m.SnapshotHash, err.Error())
+			logger.Verbosef("CosiLoop cosiHandleAction cosiHandleCommitment SendTransactionChallengeMessage(%s, %s) ERROR %v\n", id, m.SnapshotHash, err)
 		}
 	}
 	return nil
 }
 
+func (chain *Chain) cosiHandleFullChallenge(m *CosiAction) error {
+	logger.Verbosef("CosiLoop cosiHandleAction cosiHandleFullChallenge %v\n", m)
+	if m.random == nil {
+		panic(m.SnapshotHash)
+	}
+
+	s := m.Snapshot
+	if chain.IsPledging() && s.RoundNumber == 0 {
+	} else if chain.State == nil {
+		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleFullChallenge %s %v empty final round\n", m.PeerId, m.Snapshot)
+		return nil
+	} else {
+		cache, final := chain.StateCopy()
+		if s.RoundNumber < cache.Number {
+			logger.Verbosef("CosiLoop cosiHandleAction cosiHandleFullChallenge %s %v expired %d %d\n", m.PeerId, m.Snapshot, s.RoundNumber, cache.Number)
+			return nil
+		}
+		if s.RoundNumber > cache.Number+1 {
+			logger.Verbosef("CosiLoop cosiHandleAction cosiHandleFullChallenge %s %v in future %d %d\n", m.PeerId, m.Snapshot, s.RoundNumber, cache.Number)
+			return nil
+		}
+		if s.Timestamp <= final.Start+config.SnapshotRoundGap {
+			logger.Verbosef("CosiLoop cosiHandleAction cosiHandleFullChallenge %s %v invalid timestamp %d %d\n", m.PeerId, m.Snapshot, s.Timestamp, final.Start+config.SnapshotRoundGap)
+			return nil
+		}
+		if s.RoundNumber == cache.Number && !s.References.Equal(cache.References) {
+			err := chain.updateEmptyHeadRoundAndPersist(m, final, cache, s.References, s.Timestamp, true)
+			if err != nil {
+				logger.Verbosef("CosiLoop cosiHandleAction cosiHandleFullChallenge %s %v updateEmptyHeadRoundAndPersist %v\n", m.PeerId, m.Snapshot, err)
+				return nil
+			}
+			return chain.AppendCosiAction(m)
+		}
+		if s.RoundNumber == cache.Number+1 {
+			nc, nf, _, err := chain.startNewRoundAndPersist(cache, s.References, s.Timestamp, false)
+			if err != nil {
+				logger.Verbosef("CosiLoop cosiHandleAction cosiHandleFullChallenge %s %v startNewRoundAndPersist %s\n", m.PeerId, m.Snapshot, err)
+				return chain.AppendCosiAction(m)
+			} else if nf == nil {
+				logger.Verbosef("CosiLoop cosiHandleAction cosiHandleFullChallenge %s %v startNewRoundAndPersist failed\n", m.PeerId, m.Snapshot)
+				return nil
+			}
+			cache, final = nc, nf
+			chain.CosiVerifiers = make(map[crypto.Hash]*CosiVerifier)
+		}
+
+		if err := cache.ValidateSnapshot(s); err != nil {
+			logger.Verbosef("CosiLoop cosiHandleAction cosiHandleFullChallenge %s %v ValidateSnapshot %s\n", m.PeerId, m.Snapshot, err)
+			return nil
+		}
+	}
+
+	v := &CosiVerifier{Snapshot: s, Commitment: m.Commitment, random: m.random}
+	chain.CosiVerifiers[s.Hash] = v
+	chain.CosiVerifiers[s.SoleTransaction()] = v
+
+	ccm := &CosiAction{
+		PeerId:       m.PeerId,
+		Action:       CosiActionExternalChallenge,
+		SnapshotHash: s.Hash,
+		Signature:    m.Signature,
+		Transaction:  m.Transaction,
+	}
+	chain.AppendCosiAction(ccm)
+	return nil
+}
+
 func (chain *Chain) cosiHandleChallenge(m *CosiAction) error {
 	logger.Verbosef("CosiLoop cosiHandleAction cosiHandleChallenge %v\n", m)
 	v := chain.CosiVerifiers[m.SnapshotHash]
@@ -459,7 +599,7 @@ func (chain *Chain) cosiHandleChallenge(m *CosiAction) error {
 		return nil
 	}
 	if !pub.VerifyWithChallenge(m.SnapshotHash[:], sig, challenge) {
-		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleChallenge %v VerifyWithChallenge ERROR\n", m)
+		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleChallenge %v VerifyWithChallenge ERROR %v %v\n", m, sig, challenge)
 		return nil
 	}
 
@@ -471,7 +611,7 @@ func (chain *Chain) cosiHandleChallenge(m *CosiAction) error {
 	}
 	err = chain.node.Peer.SendSnapshotResponseMessage(m.PeerId, m.SnapshotHash, response)
 	if err != nil {
-		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleChallenge SendSnapshotResponseMessage(%s, %s) ERROR %s\n", m.PeerId, m.SnapshotHash, err.Error())
+		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleChallenge SendSnapshotResponseMessage(%s, %s) ERROR %v\n", m.PeerId, m.SnapshotHash, err)
 	}
 	return nil
 }
@@ -488,7 +628,7 @@ func (chain *Chain) cosiHandleResponse(m *CosiAction) error {
 		logger.Verbosef("CosiLoop cosiHandleAction cosiHandleResponse %v EXCEED\n", m)
 		return nil
 	}
-	base := chain.node.ConsensusThreshold(s.Timestamp)
+	base := chain.node.ConsensusThreshold(s.Timestamp, false)
 	agg.Responses[cd.PN.ConsensusIndex] = m.Response
 	logger.Verbosef("CosiLoop cosiHandleAction cosiHandleResponse %v NOW %d %d %d\n", m, len(agg.Responses), len(agg.Commitments), base)
 	if len(agg.Responses) != len(agg.Commitments) {
@@ -536,21 +676,21 @@ func (chain *Chain) cosiHandleResponse(m *CosiAction) error {
 		chain.AddSnapshot(final, cache, s, signers)
 	}
 
-	nodes := chain.node.NodesListWithoutState(s.Timestamp, true)
+	nodes := chain.node.cosiAcceptedNodesList(s.Timestamp)
 	for _, cn := range nodes {
 		id := cn.IdForNetwork
 		if agg.Responses[cn.ConsensusIndex] == nil {
-			err := chain.node.SendTransactionToPeer(id, s.Transaction)
+			err := chain.node.SendTransactionToPeer(id, s.SoleTransaction())
 			if err != nil {
-				logger.Verbosef("CosiLoop cosiHandleAction cosiHandleResponse SendTransactionToPeer(%s, %s) ERROR %s\n", id, m.SnapshotHash, err.Error())
+				logger.Verbosef("CosiLoop cosiHandleAction cosiHandleResponse SendTransactionToPeer(%s, %s) ERROR %v\n", id, m.SnapshotHash, err)
 			}
 		}
 		err := chain.node.Peer.SendSnapshotFinalizationMessage(id, s)
 		if err != nil {
-			logger.Verbosef("CosiLoop cosiHandleAction cosiHandleResponse SendSnapshotFinalizationMessage(%s, %s) ERROR %s\n", id, m.SnapshotHash, err.Error())
+			logger.Verbosef("CosiLoop cosiHandleAction cosiHandleResponse SendSnapshotFinalizationMessage(%s, %s) ERROR %v\n", id, m.SnapshotHash, err)
 		}
 	}
-	return chain.node.reloadConsensusNodesList(s, cd.TX)
+	return chain.node.reloadConsensusState(s, cd.TX)
 }
 
 func (chain *Chain) cosiHandleFinalization(m *CosiAction) error {
@@ -579,7 +719,7 @@ func (chain *Chain) cosiHandleFinalization(m *CosiAction) error {
 				return nil
 			}
 			if dummy {
-				logger.Verbosef("ERROR handleFinalization startNewRound DUMMY %s %s %d\n", m.PeerId, s.Hash, chain.node.ConsensusThreshold(s.Timestamp))
+				logger.Verbosef("ERROR handleFinalization startNewRound DUMMY %s %s %d\n", m.PeerId, s.Hash, chain.node.ConsensusThreshold(s.Timestamp, true))
 				return nil
 			}
 		}
@@ -587,16 +727,16 @@ func (chain *Chain) cosiHandleFinalization(m *CosiAction) error {
 
 	signers, finalized := chain.verifyFinalization(s)
 	if !finalized {
-		logger.Verbosef("ERROR handleFinalization verifyFinalization %s %v %d\n", m.PeerId, s, chain.node.ConsensusThreshold(s.Timestamp))
+		logger.Verbosef("ERROR handleFinalization verifyFinalization %s %v %d\n", m.PeerId, s, chain.node.ConsensusThreshold(s.Timestamp, true))
 		return nil
 	}
 
 	tx, _, err := chain.node.validateSnapshotTransaction(s, true)
 	if err != nil {
-		logger.Verbosef("ERROR handleFinalization checkFinalSnapshotTransaction %s %s %d %s\n", m.PeerId, s.Hash, chain.node.ConsensusThreshold(s.Timestamp), err.Error())
+		logger.Verbosef("ERROR handleFinalization checkFinalSnapshotTransaction %s %s %d %v\n", m.PeerId, s.Hash, chain.node.ConsensusThreshold(s.Timestamp, true), err)
 		return nil
 	} else if tx == nil {
-		logger.Verbosef("ERROR handleFinalization checkFinalSnapshotTransaction %s %s %d %s\n", m.PeerId, s.Hash, chain.node.ConsensusThreshold(s.Timestamp), "tx empty")
+		logger.Verbosef("ERROR handleFinalization checkFinalSnapshotTransaction %s %s %d %s\n", m.PeerId, s.Hash, chain.node.ConsensusThreshold(s.Timestamp, true), "tx empty")
 		m.WantTx = true
 		return nil
 	}
@@ -606,7 +746,7 @@ func (chain *Chain) cosiHandleFinalization(m *CosiAction) error {
 		if err != nil {
 			return err
 		}
-		return chain.node.reloadConsensusNodesList(s, tx)
+		return chain.node.reloadConsensusState(s, tx)
 	} else if chain.State == nil {
 		return nil
 	}
@@ -621,12 +761,124 @@ func (chain *Chain) cosiHandleFinalization(m *CosiAction) error {
 	}
 
 	if err := cache.ValidateSnapshot(s); err != nil {
-		logger.Verbosef("ERROR cosiHandleFinalization ValidateSnapshot %s %v %s\n", m.PeerId, s, err.Error())
+		logger.Verbosef("ERROR cosiHandleFinalization ValidateSnapshot %s %v %v\n", m.PeerId, s, err)
 		return nil
 	}
 	chain.AddSnapshot(final, cache, s, signers)
 	m.finalized = true
-	return chain.node.reloadConsensusNodesList(s, tx)
+	return chain.node.reloadConsensusState(s, tx)
+}
+
+func (chain *Chain) cosiPopCommitment(peerId crypto.Hash) *crypto.Key {
+	if chain.ChainId != chain.node.IdForNetwork {
+		panic(chain.ChainId)
+	}
+	if chain.ChainId == peerId {
+		panic(peerId)
+	}
+	commitments := chain.CosiCommitments[peerId]
+	if len(commitments) == 0 {
+		return nil
+	}
+	commitment := commitments[0]
+	chain.CosiCommitments[peerId] = commitments[1:]
+	chain.UsedCommitments[*commitment] = true
+	return commitment
+}
+
+func (chain *Chain) cosiAddCommitments(m *CosiAction) error {
+	if chain.ChainId != chain.node.IdForNetwork {
+		panic(chain.ChainId)
+	}
+	if chain.ChainId == m.PeerId {
+		panic(m.PeerId)
+	}
+	if rn := chain.node.GetRemovingOrSlashingNode(m.PeerId); rn != nil {
+		return nil
+	}
+	var commitments []*crypto.Key
+	for _, k := range m.Commitments {
+		if !chain.UsedCommitments[*k] {
+			commitments = append(commitments, k)
+		}
+	}
+	logger.Verbosef("cosiAddCommitments(%s, %d) => %d %d", m.PeerId, len(m.Commitments), len(commitments), len(chain.UsedCommitments))
+	chain.CosiCommitments[m.PeerId] = commitments
+	return nil
+}
+
+func (chain *Chain) cosiRetrieveRandom(snap crypto.Hash, peerId crypto.Hash, challenge *crypto.Key) *crypto.Key {
+	if chain.ChainId == chain.node.IdForNetwork {
+		panic(chain.ChainId)
+	}
+	if chain.ChainId != peerId {
+		panic(peerId)
+	}
+	r := chain.UsedRandoms[snap]
+	if r != nil && r.Public() == *challenge {
+		return r
+	}
+	cm := chain.CosiRandoms
+	if cm == nil {
+		return nil
+	}
+	r = cm[*challenge]
+	if r == nil {
+		return nil
+	}
+	chain.UsedRandoms[snap] = r
+	delete(chain.CosiRandoms, *challenge)
+	return r
+}
+
+func (chain *Chain) cosiPrepareRandomsAndSendCommitments(peerId crypto.Hash) error {
+	const maximum = 512
+	if chain.ChainId == chain.node.IdForNetwork {
+		panic(chain.ChainId)
+	}
+	if chain.ChainId != peerId {
+		panic(peerId)
+	}
+
+	last := chain.ComitmentsSentTime.Add(time.Duration(config.SnapshotRoundGap) * 10)
+	if last.After(clock.Now()) && len(chain.CosiRandoms) > maximum/2 {
+		return nil
+	}
+
+	// FIXME always generate new randoms, may bloat the memory
+	commitments := make([]*crypto.Key, maximum)
+	cm := make(map[crypto.Key]*crypto.Key, maximum)
+	for i := 0; i < maximum; i++ {
+		r := crypto.CosiCommit(rand.Reader)
+		k := r.Public()
+		commitments[i] = &k
+		cm[k] = r
+	}
+
+	if chain.CosiRandoms == nil {
+		chain.CosiRandoms = make(map[crypto.Key]*crypto.Key)
+	}
+	for _, r := range cm {
+		chain.CosiRandoms[r.Public()] = r
+	}
+	chain.ComitmentsSentTime = clock.Now()
+	return chain.node.Peer.SendCommitmentsMessage(peerId, commitments)
+}
+
+func (node *Node) CosiQueueExternalCommitments(peerId crypto.Hash, commitments []*crypto.Key) error {
+	logger.Debugf("CosiQueueExternalCommitments(%s, %d)\n", peerId, len(commitments))
+	if node.GetAcceptedOrPledgingNode(peerId) == nil {
+		logger.Verbosef("CosiQueueExternalCommitments(%s, %d) from malicious node\n", peerId, len(commitments))
+		return nil
+	}
+
+	m := &CosiAction{
+		PeerId:      peerId,
+		Action:      CosiActionExternalCommitments,
+		Commitments: commitments,
+	}
+	node.chain.AppendCosiAction(m)
+	return nil
 }
 
 func (node *Node) CosiQueueExternalAnnouncement(peerId crypto.Hash, s *common.Snapshot, commitment *crypto.Key) error {
@@ -635,7 +887,7 @@ func (node *Node) CosiQueueExternalAnnouncement(peerId crypto.Hash, s *common.Sn
 		logger.Verbosef("CosiQueueExternalAnnouncement(%s, %v) from malicious node\n", peerId, s)
 		return nil
 	}
-	chain := node.GetOrCreateChain(s.NodeId)
+	chain := node.getOrCreateChain(s.NodeId)
 
 	s.Hash = s.PayloadHash()
 	m := &CosiAction{
@@ -673,7 +925,7 @@ func (node *Node) CosiQueueExternalChallenge(peerId crypto.Hash, snap crypto.Has
 		logger.Verbosef("CosiQueueExternalChallenge(%s, %s) from malicious node\n", peerId, snap)
 		return nil
 	}
-	chain := node.GetOrCreateChain(peerId)
+	chain := node.getOrCreateChain(peerId)
 
 	m := &CosiAction{
 		PeerId:       peerId,
@@ -686,6 +938,29 @@ func (node *Node) CosiQueueExternalChallenge(peerId crypto.Hash, snap crypto.Has
 	return nil
 }
 
+func (node *Node) CosiQueueExternalFullChallenge(peerId crypto.Hash, s *common.Snapshot, commitment, challenge *crypto.Key, cosi *crypto.CosiSignature, ver *common.VersionedTransaction) error {
+	logger.Debugf("CosiQueueExternalFullChallenge(%s, %v)\n", peerId, s)
+	if node.GetAcceptedOrPledgingNode(peerId) == nil {
+		logger.Verbosef("CosiQueueExternalFullChallenge(%s, %v) from malicious node\n", peerId, s)
+		return nil
+	}
+	chain := node.getOrCreateChain(peerId)
+
+	s.Hash = s.PayloadHash()
+	m := &CosiAction{
+		PeerId:       peerId,
+		Action:       CosiActionExternalFullChallenge,
+		Snapshot:     s,
+		SnapshotHash: s.Hash,
+		Commitment:   commitment,
+		Challenge:    challenge,
+		Signature:    cosi,
+		Transaction:  ver,
+	}
+	chain.AppendCosiAction(m)
+	return nil
+}
+
 func (node *Node) CosiAggregateSelfResponses(peerId crypto.Hash, snap crypto.Hash, response *[32]byte) error {
 	logger.Debugf("CosiAggregateSelfResponses(%s, %s)\n", peerId, snap)
 	if node.GetAcceptedOrPledgingNode(peerId) == nil {
@@ -718,17 +993,17 @@ func (node *Node) VerifyAndQueueAppendSnapshotFinalization(peerId crypto.Hash, s
 		return nil
 	}
 
-	tx, err := node.checkTxInStorage(s.Transaction)
+	tx, err := node.checkTxInStorage(s.SoleTransaction())
 	if err != nil {
 		logger.Verbosef("VerifyAndQueueAppendSnapshotFinalization(%s, %s) check tx error %s\n", peerId, s.Hash, err)
 	} else if tx == nil {
-		logger.Verbosef("VerifyAndQueueAppendSnapshotFinalization(%s, %s) SendTransactionRequestMessage %s\n", peerId, s.Hash, s.Transaction)
-		node.Peer.SendTransactionRequestMessage(peerId, s.Transaction)
+		logger.Verbosef("VerifyAndQueueAppendSnapshotFinalization(%s, %s) SendTransactionRequestMessage %s\n", peerId, s.Hash, s.SoleTransaction())
+		node.Peer.SendTransactionRequestMessage(peerId, s.SoleTransaction())
 	}
 
-	chain := node.GetOrCreateChain(s.NodeId)
+	chain := node.getOrCreateChain(s.NodeId)
 	if _, finalized := chain.verifyFinalization(s); !finalized {
-		logger.Verbosef("ERROR VerifyAndQueueAppendSnapshotFinalization %s %v %d %t %v %v\n", peerId, s, node.ConsensusThreshold(s.Timestamp), chain.IsPledging(), chain.State, chain.ConsensusInfo)
+		logger.Verbosef("ERROR VerifyAndQueueAppendSnapshotFinalization %s %v %d %t %v %v\n", peerId, s, node.ConsensusThreshold(s.Timestamp, true), chain.IsPledging(), chain.State, chain.ConsensusInfo)
 		return nil
 	}
 
