@@ -53,6 +53,9 @@ var (
 		"0213977d3c00a91de68904fb03ce3982e139200a2ce2e6f5332c9c3fb83743c5": true,
 		"d598c36ed84b4318dffbeb81efac93be2bfd22a76f5099eef8e6a5b508628a8a": true,
 	}
+	MainnetMultiplePledgingSnapshotHackMap = map[string]bool{
+		"75d9cc79b7f80d3cf96a232b7d921d9e9c2cb93fd8df08c9ab244273f6bbd77a": true,
+	}
 )
 
 func (node *Node) ElectionLoop() {
@@ -61,7 +64,7 @@ func (node *Node) ElectionLoop() {
 	ticker := time.NewTicker(time.Duration(node.custom.Node.KernelOprationPeriod) * time.Second)
 	defer ticker.Stop()
 
-	chain := node.GetOrCreateChain(node.IdForNetwork)
+	chain := node.BootChain(node.IdForNetwork)
 	for chain.State == nil {
 		select {
 		case <-node.done:
@@ -164,7 +167,7 @@ func (node *Node) buildNodeRemoveTransaction(nodeId crypto.Hash, timestamp uint6
 		return nil, fmt.Errorf("invalid accept transaction extra %s %s %s", hex.EncodeToString(accept.Extra), signer, hex.EncodeToString(payee))
 	}
 
-	tx := common.NewTransaction(common.XINAssetId)
+	tx := node.NewTransaction(common.XINAssetId)
 	tx.AddInput(candi.Transaction, 0)
 	tx.Extra = accept.Extra
 	script := common.NewThresholdScript(1)
@@ -173,7 +176,7 @@ func (node *Node) buildNodeRemoveTransaction(nodeId crypto.Hash, timestamp uint6
 	seed := append(si[:], si[:]...)
 	tx.AddOutputWithType(common.OutputTypeNodeRemove, []*common.Address{&candi.Payee}, script, accept.Outputs[0].Amount, seed)
 
-	ver := tx.AsLatestVersion()
+	ver := tx.AsVersioned()
 	fork := uint64(ElectionTransactionV2ForkHack.UnixNano())
 	if node.networkId.String() == config.MainnetId && timestamp < fork {
 		ver.Version = 1
@@ -196,12 +199,13 @@ func (node *Node) tryToSendRemoveTransaction() error {
 	if err != nil {
 		return err
 	}
-	chain := node.GetOrCreateChain(node.IdForNetwork)
-	return chain.AppendSelfEmpty(&common.Snapshot{
-		Version:     common.SnapshotVersion,
-		NodeId:      node.IdForNetwork,
-		Transaction: tx.PayloadHash(),
-	})
+	chain := node.getOrCreateChain(node.IdForNetwork)
+	s := &common.Snapshot{
+		Version: common.SnapshotVersionCommonEncoding,
+		NodeId:  node.IdForNetwork,
+	}
+	s.AddSoleTransaction(tx.PayloadHash())
+	return chain.AppendSelfEmpty(s)
 }
 
 func (node *Node) validateNodeRemoveSnapshot(s *common.Snapshot, tx *common.VersionedTransaction) error {
@@ -294,12 +298,12 @@ func (chain *Chain) buildNodeAcceptTransaction(timestamp uint64, s *common.Snaps
 		return nil, fmt.Errorf("invalid pledge transaction extra %s %s", hex.EncodeToString(pledge.Extra[:len(signer)]), signer)
 	}
 
-	tx := common.NewTransaction(common.XINAssetId)
+	tx := chain.node.NewTransaction(common.XINAssetId)
 	tx.AddInput(ci.Transaction, 0)
 	tx.AddOutputWithType(common.OutputTypeNodeAccept, nil, common.Script{}, pledge.Outputs[0].Amount, []byte{})
 	tx.Extra = pledge.Extra
 
-	ver := tx.AsLatestVersion()
+	ver := tx.AsVersioned()
 	fork := uint64(ElectionTransactionV2ForkHack.UnixNano())
 	if chain.node.networkId.String() == config.MainnetId && timestamp < fork {
 		ver.Version = 1
@@ -322,11 +326,12 @@ func (chain *Chain) tryToSendAcceptTransaction() error {
 	if err != nil {
 		return err
 	}
-	chain.AppendSelfEmpty(&common.Snapshot{
-		Version:     common.SnapshotVersion,
-		NodeId:      chain.ChainId,
-		Transaction: ver.PayloadHash(),
-	})
+	s := &common.Snapshot{
+		Version: common.SnapshotVersionCommonEncoding,
+		NodeId:  chain.ChainId,
+	}
+	s.AddSoleTransaction(ver.PayloadHash())
+	chain.AppendSelfEmpty(s)
 	logger.Println("tryToSendAcceptTransaction", ver.PayloadHash(), hex.EncodeToString(ver.Marshal()))
 	return nil
 }
@@ -340,7 +345,7 @@ func (node *Node) validateNodeAcceptSnapshot(s *common.Snapshot, tx *common.Vers
 		return fmt.Errorf("invalid snapshot round %d", s.RoundNumber)
 	}
 
-	chain := node.GetOrCreateChain(s.NodeId)
+	chain := node.getOrCreateChain(s.NodeId)
 	ver, err := chain.buildNodeAcceptTransaction(timestamp, s, finalized)
 	if err != nil {
 		return err
@@ -352,7 +357,15 @@ func (node *Node) validateNodeAcceptSnapshot(s *common.Snapshot, tx *common.Vers
 	return nil
 }
 
-func (node *Node) reloadConsensusNodesList(s *common.Snapshot, tx *common.VersionedTransaction) error {
+func (node *Node) reloadConsensusState(s *common.Snapshot, tx *common.VersionedTransaction) error {
+	if tx.TransactionType() == common.TransactionTypeMint {
+		mint, err := node.persistStore.ReadLastMintDistribution(common.MintGroupKernelNode)
+		if err != nil {
+			return err
+		}
+		node.LastMint = mint.Batch
+		return nil
+	}
 	switch tx.TransactionType() {
 	case common.TransactionTypeNodePledge,
 		common.TransactionTypeNodeCancel,
@@ -361,13 +374,13 @@ func (node *Node) reloadConsensusNodesList(s *common.Snapshot, tx *common.Versio
 	default:
 		return nil
 	}
-	logger.Printf("reloadConsensusNodesList(%v, %v)\n", s, tx)
+	logger.Printf("reloadConsensusState(%v, %v)\n", s, tx)
 	err := node.LoadConsensusNodes()
 	if err != nil {
 		return err
 	}
 
-	chain := node.GetOrCreateChain(s.NodeId)
+	chain := node.BootChain(s.NodeId)
 	err = chain.loadState()
 	if err != nil {
 		return err
@@ -385,7 +398,7 @@ func (node *Node) reloadConsensusNodesList(s *common.Snapshot, tx *common.Versio
 		return nil
 	}
 
-	chain = node.GetOrCreateChain(id)
+	chain = node.BootChain(id)
 	err = chain.loadState()
 	if err != nil {
 		return err
@@ -406,7 +419,7 @@ func (node *Node) finalizeNodeAcceptSnapshot(s *common.Snapshot, signers []crypt
 	if err := cache.validateSnapshot(s, true); err != nil {
 		panic("should never be here")
 	}
-	err := node.persistStore.StartNewRound(cache.NodeId, cache.Number, cache.References, cache.Timestamp)
+	err := node.persistStore.StartNewRound(cache.NodeId, cache.Number, cache.References, 0)
 	if err != nil {
 		panic(err)
 	}
@@ -427,13 +440,14 @@ func (node *Node) finalizeNodeAcceptSnapshot(s *common.Snapshot, signers []crypt
 			External: external.Hash,
 		},
 	}
-	err = node.persistStore.StartNewRound(cache.NodeId, cache.Number, cache.References, cache.Timestamp)
+	err = node.persistStore.StartNewRound(cache.NodeId, cache.Number, cache.References, final.Start)
 	if err != nil {
 		panic(err)
 	}
 
-	chain := node.GetOrCreateChain(s.NodeId)
-	if err := chain.loadState(); err != nil {
+	chain := node.BootChain(s.NodeId)
+	err = chain.loadState()
+	if err != nil {
 		return err
 	}
 	chain.StepForward()
@@ -482,6 +496,9 @@ func (node *Node) validateNodePledgeSnapshot(s *common.Snapshot, tx *common.Vers
 		if cn.Transaction == tx.PayloadHash() {
 			continue
 		}
+		if MainnetMultiplePledgingSnapshotHackMap[tx.PayloadHash().String()] {
+			continue
+		}
 		if timestamp < cn.Timestamp {
 			return fmt.Errorf("invalid snapshot timestamp %d %d", cn.Timestamp, timestamp)
 		}
