@@ -21,150 +21,31 @@ import (
 	"github.com/MixinNetwork/mixin/kernel"
 	"github.com/MixinNetwork/mixin/logger"
 	"github.com/MixinNetwork/mixin/storage"
-	"github.com/VictoriaMetrics/fastcache"
+	"github.com/dgraph-io/ristretto"
 	"github.com/stretchr/testify/assert"
 )
 
-const (
+var (
 	NODES  = 8
 	INPUTS = 100
 )
 
-func TestAllTransactionsToSingleGenesisNode(t *testing.T) {
+func TestConsensus(t *testing.T) {
+	testConsensus(t, 0)
+}
+
+func testConsensus(t *testing.T, snapVersionMint int) {
 	assert := assert.New(t)
-
-	kernel.TestMockReset()
-
-	root, err := os.MkdirTemp("", "mixin-attsg-test")
-	assert.Nil(err)
-	defer os.RemoveAll(root)
-
-	accounts, _, _, _ := setupTestNet(root)
-	assert.Len(accounts, NODES)
-
-	epoch := time.Unix(1551312000, 0)
-	nodes := make([]*Node, 0)
-	instances := make([]*kernel.Node, 0)
-	for i := range accounts {
-		dir := fmt.Sprintf("%s/mixin-170%02d", root, i+1)
-		custom, err := config.Initialize(dir + "/config.toml")
-		assert.Nil(err)
-		cache := fastcache.New(custom.Node.MemoryCacheSize * 1024 * 1024)
-		store, err := storage.NewBadgerStore(custom, dir)
-		assert.Nil(err)
-		assert.NotNil(store)
-		if i == 0 {
-			kernel.TestMockDiff(epoch.Sub(time.Now()))
-		}
-		node, err := kernel.SetupNode(custom, store, cache, fmt.Sprintf(":170%02d", i+1), dir)
-		assert.Nil(err)
-		assert.NotNil(node)
-		err = node.PingNeighborsFromConfig()
-		assert.Nil(err)
-		instances = append(instances, node)
-		host := fmt.Sprintf("127.0.0.1:180%02d", i+1)
-		nodes = append(nodes, &Node{Signer: node.Signer, Host: host})
-
-		server := NewServer(custom, store, node, 18000+i+1)
-		defer server.Close()
-		go func(node *kernel.Node, store storage.Store, num int, s *http.Server) {
-			go s.ListenAndServe()
-			go node.Loop()
-		}(node, store, i, server)
-	}
-	defer func() {
-		var wg sync.WaitGroup
-		for _, n := range instances {
-			wg.Add(1)
-			go func(node *kernel.Node) {
-				node.Teardown()
-				wg.Done()
-			}(n)
-		}
-		wg.Wait()
-	}()
-	time.Sleep(3 * time.Second)
-
-	rand.Seed(time.Now().UnixNano())
-	target := nodes[rand.Intn(len(nodes))]
-
-	tl, sl := testVerifySnapshots(assert, nodes)
-	assert.Equal(NODES+1, len(tl))
-	assert.Equal(NODES+1, len(sl))
-	gt := testVerifyInfo(assert, nodes)
-	assert.Truef(gt.Timestamp.Before(epoch.Add(1*time.Second)), "%s should before %s", gt.Timestamp, epoch.Add(1*time.Second))
-
-	genesisAmount := float64(10003.5) / INPUTS
-	domainAddress := accounts[0].String()
-	deposits := make([]*common.VersionedTransaction, 0)
-	for i := 0; i < INPUTS; i++ {
-		raw := fmt.Sprintf(`{"version":2,"asset":"a99c2e0e2b1da4d648755ef19bd95139acbbe6564cfb06dec7cd34931ca72cdc","inputs":[{"deposit":{"chain":"8dd50817c082cdcdd6f167514928767a4b52426997bd6d4930eca101c5ff8a27","asset":"0xa974c709cfb4566686553a20790685a47aceaa33","transaction":"0xc7c1132b58e1f64c263957d7857fe5ec5294fce95d30dcd64efef71da1%06d","index":0,"amount":"%f"}}],"outputs":[{"type":0,"amount":"%f","script":"fffe01","accounts":["%s"]}]}`, i, genesisAmount, genesisAmount, domainAddress)
-		tx, err := testSignTransaction(target.Host, accounts[0], raw)
-		assert.Nil(err)
-		assert.NotNil(tx)
-		deposits = append(deposits, &common.VersionedTransaction{SignedTransaction: *tx})
-	}
-
-	for _, d := range deposits {
-		n, raw := target.Host, hex.EncodeToString(d.Marshal())
-		id, err := testSendTransaction(n, raw)
-		assert.Nil(err)
-		assert.Len(id, 75)
-	}
-
-	time.Sleep(3 * time.Second)
-	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS+NODES+1, len(tl))
-	gt = testVerifyInfo(assert, nodes)
-	assert.Truef(gt.Timestamp.Before(epoch.Add(1*time.Second)), "%s should before %s", gt.Timestamp, epoch.Add(1*time.Second))
-
-	utxos := make([]*common.VersionedTransaction, 0)
-	for _, d := range deposits {
-		raw := fmt.Sprintf(`{"version":2,"asset":"a99c2e0e2b1da4d648755ef19bd95139acbbe6564cfb06dec7cd34931ca72cdc","inputs":[{"hash":"%s","index":0}],"outputs":[{"type":0,"amount":"%f","script":"fffe01","accounts":["%s"]}]}`, d.PayloadHash().String(), genesisAmount, domainAddress)
-		rand.Seed(time.Now().UnixNano())
-		tx, err := testSignTransaction(nodes[0].Host, accounts[0], raw)
-		assert.Nil(err)
-		assert.NotNil(tx)
-		if tx != nil {
-			utxos = append(utxos, &common.VersionedTransaction{SignedTransaction: *tx})
-		}
-	}
-	assert.Equal(INPUTS, len(utxos))
-
-	for _, tx := range utxos {
-		n, raw := nodes[0].Host, hex.EncodeToString(tx.Marshal())
-		id, err := testSendTransaction(n, raw)
-		assert.Nil(err)
-		assert.Len(id, 75)
-	}
-
-	time.Sleep(3 * time.Second)
-	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1, len(tl))
-	gt = testVerifyInfo(assert, nodes)
-	assert.True(gt.Timestamp.Before(epoch.Add(31 * time.Second)))
-}
-
-func TestConsensusSingle(t *testing.T) {
-	testConsensus(t, 1)
-}
-
-func TestConsensusDouble(t *testing.T) {
-	testConsensus(t, 2)
-}
-
-func TestConsensusMany(t *testing.T) {
-	testConsensus(t, -1)
-}
-
-func testConsensus(t *testing.T, dup int) {
-	assert := assert.New(t)
-
 	kernel.TestMockReset()
 
 	level, _ := strconv.ParseInt(os.Getenv("LOG"), 10, 64)
+	enableElection, _ := strconv.ParseBool(os.Getenv("ELECTION"))
+	inputs, _ := strconv.ParseInt(os.Getenv("INPUT"), 10, 64)
 	logger.SetLevel(int(level))
-	logger.SetLimiter(10)
+	if inputs > 0 {
+		INPUTS = int(inputs)
+	}
+	t.Logf("TEST WITH %d INPUTS AT %s\n", INPUTS, time.Now())
 
 	root, err := os.MkdirTemp("", "mixin-consensus-test")
 	assert.Nil(err)
@@ -180,7 +61,7 @@ func testConsensus(t *testing.T, dup int) {
 		dir := fmt.Sprintf("%s/mixin-170%02d", root, i+1)
 		custom, err := config.Initialize(dir + "/config.toml")
 		assert.Nil(err)
-		cache := fastcache.New(custom.Node.MemoryCacheSize * 1024 * 1024)
+		cache := newCache(custom)
 		store, err := storage.NewBadgerStore(custom, dir)
 		assert.Nil(err)
 		assert.NotNil(store)
@@ -195,7 +76,7 @@ func testConsensus(t *testing.T, dup int) {
 		instances = append(instances, node)
 		host := fmt.Sprintf("127.0.0.1:180%02d", i+1)
 		nodes = append(nodes, &Node{Signer: node.Signer, Host: host})
-		t.Logf("NODES#%d %s\n", i, node.IdForNetwork)
+		t.Logf("NODES#%d %s %s\n", i, node.IdForNetwork, host)
 
 		server := NewServer(custom, store, node, 18000+i+1)
 		defer server.Close()
@@ -217,67 +98,43 @@ func testConsensus(t *testing.T, dup int) {
 	}()
 	time.Sleep(3 * time.Second)
 
+	transactionsCount := NODES + 1
 	tl, sl := testVerifySnapshots(assert, nodes)
-	assert.Equal(NODES+1, len(tl))
-	assert.Equal(NODES+1, len(sl))
+	assert.Equal(transactionsCount, len(tl))
+	assert.Equal(transactionsCount, len(sl))
 	gt := testVerifyInfo(assert, nodes)
 	assert.Truef(gt.Timestamp.Before(epoch.Add(1*time.Second)), "%s should before %s", gt.Timestamp, epoch.Add(1*time.Second))
 
-	genesisAmount := float64(10003.5) / INPUTS
+	genesisAmount := 10003.5 / float64(INPUTS)
 	domainAddress := accounts[0].String()
 	deposits := make([]*common.VersionedTransaction, 0)
 	for i := 0; i < INPUTS; i++ {
 		raw := fmt.Sprintf(`{"version":2,"asset":"a99c2e0e2b1da4d648755ef19bd95139acbbe6564cfb06dec7cd34931ca72cdc","inputs":[{"deposit":{"chain":"8dd50817c082cdcdd6f167514928767a4b52426997bd6d4930eca101c5ff8a27","asset":"0xa974c709cfb4566686553a20790685a47aceaa33","transaction":"0xc7c1132b58e1f64c263957d7857fe5ec5294fce95d30dcd64efef71da1%06d","index":0,"amount":"%f"}}],"outputs":[{"type":0,"amount":"%f","script":"fffe01","accounts":["%s"]}]}`, i, genesisAmount, genesisAmount, domainAddress)
 		rand.Seed(time.Now().UnixNano())
-		tx, err := testSignTransaction(nodes[rand.Intn(len(nodes))].Host, accounts[0], raw)
+		tx, err := testSignTransaction(nodes[rand.Intn(len(nodes))].Host, accounts[0], raw, snapVersionMint)
 		assert.Nil(err)
 		assert.NotNil(tx)
 		deposits = append(deposits, &common.VersionedTransaction{SignedTransaction: *tx})
 	}
 
-	for _, d := range deposits {
-		rand.Seed(time.Now().UnixNano())
-		for n := len(nodes); n > 0; n-- {
-			randIndex := rand.Intn(n)
-			nodes[n-1], nodes[randIndex] = nodes[randIndex], nodes[n-1]
-		}
-		wg := &sync.WaitGroup{}
-		start := rand.Intn(len(nodes) - 1)
-		for i := start; i < len(nodes) && i != start+dup; i++ {
-			wg.Add(1)
-			go func(n string, raw string) {
-				defer wg.Done()
-				id, err := testSendTransaction(n, raw)
-				assert.Nil(err)
-				assert.Len(id, 75)
-			}(nodes[i].Host, hex.EncodeToString(d.Marshal()))
-		}
-		wg.Wait()
-	}
-
-	time.Sleep(3 * time.Second)
+	testSendTransactionsToNodesWithRetry(t, nodes, deposits[:INPUTS/2])
+	testSendTransactionsToNodesWithRetry(t, nodes, deposits[INPUTS/2:])
+	transactionsCount = transactionsCount + INPUTS
 	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS+NODES+1, len(tl))
-	for i, d := range deposits {
-		if !tl[d.PayloadHash().String()] {
-			t.Logf("DEPOSIT MISSING %d %s\n", i, d.PayloadHash())
-			id, err := testSendTransaction(nodes[0].Host, hex.EncodeToString(d.Marshal()))
-			assert.Nil(err)
-			assert.Contains(id, d.PayloadHash().String())
-		}
-	}
+	assert.Equal(transactionsCount, len(tl))
+
 	gt = testVerifyInfo(assert, nodes)
-	assert.Truef(gt.Timestamp.Before(epoch.Add(1*time.Second)), "%s should before %s", gt.Timestamp, epoch.Add(1*time.Second))
+	assert.Truef(gt.Timestamp.Before(epoch.Add(7*time.Second)), "%s should before %s", gt.Timestamp, epoch.Add(7*time.Second))
 	hr := testDumpGraphHead(nodes[0].Host, instances[0].IdForNetwork)
 	assert.NotNil(hr)
 	assert.GreaterOrEqual(hr.Round, uint64(0))
-	t.Log("DEPOSIT TEST DONE", time.Now())
+	t.Logf("DEPOSIT TEST DONE AT %s\n", time.Now())
 
 	utxos := make([]*common.VersionedTransaction, 0)
 	for _, d := range deposits {
 		raw := fmt.Sprintf(`{"version":2,"asset":"a99c2e0e2b1da4d648755ef19bd95139acbbe6564cfb06dec7cd34931ca72cdc","inputs":[{"hash":"%s","index":0}],"outputs":[{"type":0,"amount":"%f","script":"fffe01","accounts":["%s"]}]}`, d.PayloadHash().String(), genesisAmount, domainAddress)
 		rand.Seed(time.Now().UnixNano())
-		tx, err := testSignTransaction(nodes[rand.Intn(len(nodes))].Host, accounts[0], raw)
+		tx, err := testSignTransaction(nodes[rand.Intn(len(nodes))].Host, accounts[0], raw, snapVersionMint)
 		assert.Nil(err)
 		assert.NotNil(tx)
 		if tx != nil {
@@ -286,70 +143,68 @@ func testConsensus(t *testing.T, dup int) {
 	}
 	assert.Equal(INPUTS, len(utxos))
 
-	for _, tx := range utxos {
-		rand.Seed(time.Now().UnixNano())
-		for n := len(nodes); n > 0; n-- {
-			randIndex := rand.Intn(n)
-			nodes[n-1], nodes[randIndex] = nodes[randIndex], nodes[n-1]
-		}
-		wg := &sync.WaitGroup{}
-		start := rand.Intn(len(nodes) - 1)
-		for i := start; i < len(nodes) && i != start+dup; i++ {
-			wg.Add(1)
-			go func(n string, raw string) {
-				defer wg.Done()
-				id, err := testSendTransaction(n, raw)
-				assert.Nil(err)
-				assert.Len(id, 75)
-			}(nodes[i].Host, hex.EncodeToString(tx.Marshal()))
-		}
-		wg.Wait()
-	}
-
-	time.Sleep(3 * time.Second)
+	testSendTransactionsToNodesWithRetry(t, nodes, utxos[:INPUTS/2])
+	testSendTransactionsToNodesWithRetry(t, nodes, utxos[INPUTS/2:])
+	transactionsCount = transactionsCount + INPUTS
 	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1, len(tl))
-	for i, tx := range utxos {
-		if !tl[tx.PayloadHash().String()] {
-			t.Logf("UTXO MISSING %d %s\n", i, tx.PayloadHash())
-			id, err := testSendTransaction(nodes[0].Host, hex.EncodeToString(tx.Marshal()))
-			assert.Nil(err)
-			assert.Contains(id, tx.PayloadHash().String())
-		}
-	}
+	assert.Equal(transactionsCount, len(tl))
+
 	gt = testVerifyInfo(assert, nodes)
 	assert.True(gt.Timestamp.Before(epoch.Add(31 * time.Second)))
 	hr = testDumpGraphHead(nodes[0].Host, instances[0].IdForNetwork)
 	assert.NotNil(hr)
 	assert.Greater(hr.Round, uint64(0))
-	t.Log("INPUT TEST DONE", time.Now())
+	t.Logf("INPUT TEST DONE AT %s\n", time.Now())
 
-	kernel.TestMockDiff((config.KernelMintTimeBegin + 24) * time.Hour)
-	time.Sleep(3 * time.Second)
-	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1, len(tl))
-	gt = testVerifyInfo(assert, nodes)
-	assert.True(gt.Timestamp.Before(epoch.Add(31 * time.Second)))
+	if !enableElection {
+		return
+	}
 
 	all := testListNodes(nodes[0].Host)
 	assert.Len(all, NODES)
 	assert.Equal("ACCEPTED", all[NODES-1].State)
 
-	input, err := testBuildPledgeInput(assert, nodes[0].Host, accounts[0], utxos)
-	assert.Nil(err)
-	time.Sleep(5 * time.Second)
+	input, _ := testBuildPledgeInput(t, nodes, accounts[0], utxos, snapVersionMint)
+	time.Sleep(3 * time.Second)
+	transactionsCount = transactionsCount + 1
 	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1, len(tl))
+	assert.Equal(transactionsCount, len(tl))
 	gt = testVerifyInfo(assert, nodes)
 	assert.True(gt.Timestamp.Before(epoch.Add(61 * time.Second)))
 	t.Logf("PLEDGE %s\n", input)
 
-	pn, pi, sv := testPledgeNewNode(assert, nodes[0].Host, accounts[0], gdata, plist, input, root)
+	dummyAmount := common.NewIntegerFromString("3.5").Div(NODES).String()
+	dummyInputs := make([]*common.Input, NODES)
+	for i := range dummyInputs {
+		hash, _ := crypto.HashFromString(input)
+		dummyInputs[i] = &common.Input{Hash: hash, Index: i}
+	}
+
+	for i := 0; i < 3; i++ {
+		dummyInputs = testSendDummyTransactionsWithRetry(t, nodes, accounts[0], dummyInputs, dummyAmount, snapVersionMint)
+		transactionsCount = transactionsCount + len(dummyInputs)
+	}
+
+	kernel.TestMockDiff((config.KernelMintTimeBegin + 24) * time.Hour)
+	tl, _ = testVerifySnapshots(assert, nodes)
+	assert.Equal(transactionsCount, len(tl))
+	gt = testVerifyInfo(assert, nodes)
+	assert.True(gt.Timestamp.Before(epoch.Add(31 * time.Second)))
+
+	pn, pi, sv := testPledgeNewNode(t, nodes, accounts[0], gdata, plist, input, root, snapVersionMint)
+	t.Logf("PLEDGE %s\n", pn.Signer)
+	transactionsCount = transactionsCount + 1
 	defer pi.Teardown()
 	defer sv.Close()
-	time.Sleep(3 * time.Second)
+
+	for i := 0; i < 5; i++ {
+		dummyInputs = testSendDummyTransactionsWithRetry(t, nodes, accounts[0], dummyInputs, dummyAmount, snapVersionMint)
+		transactionsCount = transactionsCount + len(dummyInputs)
+	}
+
+	transactionsCount = transactionsCount + 1
 	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1+2, len(tl))
+	assert.Equal(transactionsCount, len(tl))
 	gt = testVerifyInfo(assert, nodes)
 	assert.True(gt.Timestamp.After(epoch.Add((config.KernelMintTimeBegin + 24) * time.Hour)))
 	assert.Equal("499876.71232883", gt.PoolSize.String())
@@ -364,7 +219,7 @@ func testConsensus(t *testing.T, dup int) {
 	assert.Equal(all[NODES].Signer.String(), pn.Signer.String())
 	assert.Equal(all[NODES].Payee.String(), pn.Payee.String())
 	assert.Equal("PLEDGING", all[NODES].State)
-	t.Log("PLEDGE TEST DONE", time.Now())
+	t.Logf("PLEDGE TEST DONE AT %s\n", time.Now())
 
 	kernel.TestMockDiff(11 * time.Hour)
 	time.Sleep(3 * time.Second)
@@ -398,22 +253,18 @@ func testConsensus(t *testing.T, dup int) {
 	assert.NotNil(hr)
 	assert.Equal(uint64(0), hr.Round)
 
+	transactionsCount = transactionsCount + 1
 	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1+2+1, len(tl))
+	assert.Equal(transactionsCount, len(tl))
 	gt = testVerifyInfo(assert, nodes)
 	assert.True(gt.Timestamp.After(epoch.Add((config.KernelMintTimeBegin + 24) * time.Hour)))
 	assert.Equal("499876.71232883", gt.PoolSize.String())
+	t.Logf("ACCEPT TEST DONE AT %s\n", time.Now())
 
 	kernel.TestMockDiff(24 * time.Hour)
 	time.Sleep(3 * time.Second)
 	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1+2+1, len(tl))
-
-	input = testSendDummyTransaction(assert, nodes[0].Host, accounts[0], input, "3.5")
-	t.Logf("DUMMY 1 %s\n", input)
-	time.Sleep(3 * time.Second)
-	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1+2+1+1, len(tl))
+	assert.Equal(transactionsCount, len(tl))
 	for i := range nodes {
 		all = testListNodes(nodes[i].Host)
 		assert.Len(all, NODES+1)
@@ -421,19 +272,20 @@ func testConsensus(t *testing.T, dup int) {
 		assert.Equal(all[NODES].Payee.String(), pn.Payee.String())
 		assert.Equal("ACCEPTED", all[NODES].State)
 	}
-	t.Log("ACCEPT TEST DONE", time.Now())
-	nodes = append(nodes, &Node{Host: "127.0.0.1:18099"})
 
+	nodes = append(nodes, &Node{Host: "127.0.0.1:18099"})
 	signer, payee := testGetNodeToRemove(instances[0].NetworkId(), accounts, payees, 0)
 	assert.Equal("XINGmuYCB65rzMgUf1W35pbhj4C7fY9JrzWCL5vGRdL84SPcWVPhtBJ7DAarc1QPt564JwbEdNCH8359kdPRH1ieSM9f96RZ", signer.String())
 	assert.Equal("XINMeKsKkSJJCgLWKvakEHaXBNPGfF7RmBu9jx5VZLE6UTuEaW4wSEqVybkH4xhQcqkT5jdiguiN3B3NKt8QBZTUbqZXJ1Fq", payee.String())
-	input = testSendDummyTransaction(assert, nodes[0].Host, accounts[0], input, "3.5")
-	t.Logf("DUMMY 2 %s\n", input)
-	assert.Len(input, 64)
 	nodes = testRemoveNode(nodes, signer)
-	time.Sleep(5 * time.Second)
+	for i := 0; i < 3; i++ {
+		dummyInputs = testSendDummyTransactionsWithRetry(t, nodes, accounts[0], dummyInputs, dummyAmount, snapVersionMint)
+		transactionsCount = transactionsCount + len(dummyInputs)
+	}
+	transactionsCount = transactionsCount + 1
+
 	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1+2+1+1+2, len(tl))
+	assert.Equal(transactionsCount, len(tl))
 	for i := range nodes {
 		all = testListNodes(nodes[i].Host)
 		assert.Len(all, NODES+1)
@@ -441,35 +293,25 @@ func testConsensus(t *testing.T, dup int) {
 		assert.Equal(all[NODES].Payee.String(), payee.String())
 		assert.Equal("REMOVED", all[NODES].State)
 	}
+
 	hr = testDumpGraphHead(nodes[0].Host, instances[0].IdForNetwork)
-	assert.NotNil(hr)
-	assert.Greater(hr.Round, uint64(0))
+	assert.Greater(hr.Round, uint64(1))
 	hr = testDumpGraphHead(nodes[len(nodes)-1].Host, instances[0].IdForNetwork)
-	assert.NotNil(hr)
-	assert.Greater(hr.Round, uint64(0))
+	assert.Greater(hr.Round, uint64(1))
 	hr = testDumpGraphHead(nodes[0].Host, pi.IdForNetwork)
-	assert.NotNil(hr)
-	assert.Equal(uint64(0), hr.Round)
+	assert.Greater(hr.Round, uint64(1))
 	hr = testDumpGraphHead(nodes[len(nodes)-1].Host, pi.IdForNetwork)
-	assert.NotNil(hr)
-	assert.Equal(uint64(0), hr.Round)
+	assert.Greater(hr.Round, uint64(1))
 	hr = testDumpGraphHead(nodes[0].Host, signer.Hash().ForNetwork(instances[0].NetworkId()))
-	assert.NotNil(hr)
-	assert.Greater(hr.Round, uint64(0))
+	assert.Greater(hr.Round, uint64(1))
 	hr = testDumpGraphHead(nodes[len(nodes)-1].Host, signer.Hash().ForNetwork(instances[0].NetworkId()))
-	assert.NotNil(hr)
-	assert.Greater(hr.Round, uint64(0))
+	assert.Greater(hr.Round, uint64(1))
 
-	kernel.TestMockDiff(24 * time.Hour)
-	time.Sleep(3 * time.Second)
+	removalInputs := []*common.Input{{Hash: all[NODES].Transaction, Index: 0}}
+	removalInputs = testSendDummyTransactionsWithRetry(t, nodes[:1], payee, removalInputs, "10000", snapVersionMint)
+	transactionsCount = transactionsCount + 1
 	tl, _ = testVerifySnapshots(assert, nodes)
-
-	removal := testSendDummyTransaction(assert, nodes[0].Host, payee, all[NODES].Transaction.String(), "10000")
-	t.Logf("DUMMY 3 %s\n", removal)
-	assert.Len(removal, 64)
-	time.Sleep(3 * time.Second)
-	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1+2+1+1+2+1, len(tl))
+	assert.Equal(transactionsCount, len(tl))
 	for i := range nodes {
 		all = testListNodes(nodes[i].Host)
 		assert.Len(all, NODES+1)
@@ -477,59 +319,11 @@ func testConsensus(t *testing.T, dup int) {
 		assert.Equal(all[NODES].Payee.String(), payee.String())
 		assert.Equal("REMOVED", all[NODES].State)
 	}
+	t.Logf("REMOVE TEST DONE AT %s\n", time.Now())
 
-	signer, payee = testGetNodeToRemove(instances[0].NetworkId(), accounts, payees, 1)
-	assert.Equal("XINEBDHLoH3eDiicxWdbF7B79h1kZMT2rmuoQcvSPDYztBFvkgv8dESeCKW1Ejh8hQ7QUK2Lxd8aHa5heXmWYk2gc78wWVs", signer.String())
-	assert.Equal("XINRqCuT9qD7qd9Z4bJFhAHq4eRa1GWdeD6cM6AegbS9chCFRFpxfxvFey6LouzRDYJgKXd45YXM1douCFvwELZmQ9gSCjPP", payee.String())
-	input = testSendDummyTransaction(assert, nodes[0].Host, accounts[0], input, "3.5")
-	t.Logf("DUMMY 4 %s\n", input)
-	assert.Len(input, 64)
-	nodes = testRemoveNode(nodes, signer)
-	assert.Len(nodes, NODES+1-2)
-	time.Sleep(5 * time.Second)
-	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1+2+1+1+2+1+2, len(tl))
-	for i := range nodes {
-		all = testListNodes(nodes[i].Host)
-		assert.Len(all, NODES+1)
-		assert.Equal(all[NODES].Signer.String(), signer.String())
-		assert.Equal(all[NODES].Payee.String(), payee.String())
-		assert.Equal("REMOVED", all[NODES].State)
+	for _, node := range instances {
+		t.Log(node.IdForNetwork, node.Peer.Metric())
 	}
-
-	kernel.TestMockDiff(24 * time.Hour)
-	time.Sleep(3 * time.Second)
-	tl, _ = testVerifySnapshots(assert, nodes)
-
-	removal = testSendDummyTransaction(assert, nodes[0].Host, payee, all[NODES].Transaction.String(), "10000")
-	t.Logf("DUMMY 5 %s\n", removal)
-	assert.Len(removal, 64)
-	time.Sleep(3 * time.Second)
-	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1+2+1+1+2+1+2+1, len(tl))
-	for i := range nodes {
-		all = testListNodes(nodes[i].Host)
-		assert.Len(all, NODES+1)
-		assert.Equal(all[NODES].Signer.String(), signer.String())
-		assert.Equal(all[NODES].Payee.String(), payee.String())
-		assert.Equal("REMOVED", all[NODES].State)
-	}
-
-	input = testSendDummyTransaction(assert, nodes[0].Host, accounts[0], input, "3.5")
-	t.Logf("DUMMY 6 %s\n", input)
-	assert.Len(input, 64)
-	nodes = testRemoveNode(nodes, signer)
-	time.Sleep(5 * time.Second)
-	tl, _ = testVerifySnapshots(assert, nodes)
-	assert.Equal(INPUTS*2+NODES+1+1+2+1+1+2+1+2+1+1, len(tl))
-	for i := range nodes {
-		all = testListNodes(nodes[i].Host)
-		assert.Len(all, NODES+1)
-		assert.Equal(all[NODES].Signer.String(), signer.String())
-		assert.Equal(all[NODES].Payee.String(), payee.String())
-		assert.Equal("REMOVED", all[NODES].State)
-	}
-	t.Log("REMOVE TEST DONE", time.Now())
 }
 
 func testRemoveNode(nodes []*Node, r common.Address) []*Node {
@@ -547,48 +341,88 @@ func testRemoveNode(nodes []*Node, r common.Address) []*Node {
 	return tmp
 }
 
-func testSendDummyTransaction(assert *assert.Assertions, node string, domain common.Address, th, amount string) string {
-	raw, _ := json.Marshal(map[string]interface{}{
-		"version": 2,
-		"asset":   "a99c2e0e2b1da4d648755ef19bd95139acbbe6564cfb06dec7cd34931ca72cdc",
-		"inputs": []map[string]interface{}{{
-			"hash":  th,
-			"index": 0,
-		}},
-		"outputs": []map[string]interface{}{{
-			"type":     0,
-			"amount":   amount,
-			"script":   "fffe01",
-			"accounts": []string{domain.String()},
-		}},
-	})
-	tx, err := testSignTransaction(node, domain, string(raw))
-	assert.Nil(err)
-	ver := common.VersionedTransaction{SignedTransaction: *tx}
-	input, err := testSendTransaction(node, hex.EncodeToString(ver.Marshal()))
-	assert.Nil(err)
-	var hash map[string]string
-	err = json.Unmarshal([]byte(input), &hash)
-	assert.Nil(err)
-	return hash["hash"]
+func testSendDummyTransactionsWithRetry(t *testing.T, nodes []*Node, domain common.Address, inputs []*common.Input, amount string, snapVersionMint int) []*common.Input {
+	outputs := testSendDummyTransactions(nodes, domain, inputs, amount, snapVersionMint)
+	time.Sleep(3 * time.Second)
+
+	var missingInputs []*common.Input
+	var missingNodes []*Node
+	for i, in := range outputs {
+		data, _ := callRPC(nodes[i].Host, "gettransaction", []any{in.Hash.String()})
+		var res map[string]string
+		json.Unmarshal([]byte(data), &res)
+		hash, _ := crypto.HashFromString(res["snapshot"])
+		if hash.HasValue() {
+			continue
+		}
+		t.Logf("DUMMY UTXO %s PENDING IN %s AT %s\n", inputs[i].Hash, nodes[i].Host, time.Now())
+		hash, _ = crypto.HashFromString(res["hash"])
+		if !hash.HasValue() {
+			t.Logf("DUMMY UTXO %s MISSING IN %s AT %s\n", inputs[i].Hash, nodes[i].Host, time.Now())
+		}
+		missingInputs = append(missingInputs, inputs[i])
+		missingNodes = append(missingNodes, nodes[i])
+	}
+	if len(missingInputs) > 0 {
+		testSendDummyTransactionsWithRetry(t, missingNodes, domain, missingInputs, amount, snapVersionMint)
+	}
+	return outputs
+}
+
+func testSendDummyTransactions(nodes []*Node, domain common.Address, inputs []*common.Input, amount string, snapVersionMint int) []*common.Input {
+	outputs := make([]*common.Input, len(inputs))
+
+	var wg sync.WaitGroup
+	for i, node := range nodes {
+		wg.Add(1)
+		go func(i int, node *Node) {
+			raw, _ := json.Marshal(map[string]any{
+				"version": 2,
+				"asset":   "a99c2e0e2b1da4d648755ef19bd95139acbbe6564cfb06dec7cd34931ca72cdc",
+				"inputs": []map[string]any{{
+					"hash":  inputs[i].Hash.String(),
+					"index": inputs[i].Index,
+				}},
+				"outputs": []map[string]any{{
+					"type":     0,
+					"amount":   amount,
+					"script":   "fffe01",
+					"accounts": []string{domain.String()},
+				}},
+			})
+			tx, _ := testSignTransaction(node.Host, domain, string(raw), snapVersionMint)
+			ver := common.VersionedTransaction{SignedTransaction: *tx}
+			id, _ := testSendTransaction(node.Host, hex.EncodeToString(ver.Marshal()))
+			var res map[string]string
+			json.Unmarshal([]byte(id), &res)
+			hash, _ := crypto.HashFromString(res["hash"])
+			outputs[i] = &common.Input{Index: 0, Hash: hash}
+			wg.Done()
+		}(i, node)
+	}
+	wg.Wait()
+
+	return outputs
 }
 
 const configDataTmpl = `[node]
 signer-key = "%s"
 consensus-only = false
 memory-cache-size = 128
-kernel-operation-period = 1
+kernel-operation-period = 3
 cache-ttl = 3600
 [network]
 listener = "%s"
+metric = true
 peers = [%s]
 `
 
-func testPledgeNewNode(assert *assert.Assertions, node string, domain common.Address, genesisData []byte, plist, input, root string) (Node, *kernel.Node, *http.Server) {
+func testPledgeNewNode(t *testing.T, nodes []*Node, domain common.Address, genesisData []byte, plist, input, root string, snapVersionMint int) (Node, *kernel.Node, *http.Server) {
+	assert := assert.New(t)
 	var signer, payee common.Address
 
-	signer = testDeterminAccountByIndex(NODES, "SIGNER")
-	payee = testDeterminAccountByIndex(NODES, "PAYEE")
+	signer = testDetermineAccountByIndex(NODES, "SIGNER")
+	payee = testDetermineAccountByIndex(NODES, "PAYEE")
 
 	dir := fmt.Sprintf("%s/mixin-17099", root)
 	err := os.MkdirAll(dir, 0755)
@@ -606,28 +440,27 @@ func testPledgeNewNode(assert *assert.Assertions, node string, domain common.Add
 		panic(err)
 	}
 
-	raw, _ := json.Marshal(map[string]interface{}{
+	raw, _ := json.Marshal(map[string]any{
 		"version": 2,
 		"asset":   "a99c2e0e2b1da4d648755ef19bd95139acbbe6564cfb06dec7cd34931ca72cdc",
-		"inputs": []map[string]interface{}{{
+		"inputs": []map[string]any{{
 			"hash":  input,
-			"index": 1,
+			"index": NODES,
 		}},
-		"outputs": []map[string]interface{}{{
+		"outputs": []map[string]any{{
 			"type":   common.OutputTypeNodePledge,
 			"amount": "10000",
 		}},
 		"extra": signer.PublicSpendKey.String() + payee.PublicSpendKey.String(),
 	})
-	tx, err := testSignTransaction(node, domain, string(raw))
+	tx, err := testSignTransaction(nodes[0].Host, domain, string(raw), snapVersionMint)
 	assert.Nil(err)
 	ver := common.VersionedTransaction{SignedTransaction: *tx}
-	_, err = testSendTransaction(node, hex.EncodeToString(ver.Marshal()))
-	assert.Nil(err)
+	testSendTransactionsToNodesWithRetry(t, nodes, []*common.VersionedTransaction{&ver})
 
 	custom, err := config.Initialize(dir + "/config.toml")
 	assert.Nil(err)
-	cache := fastcache.New(custom.Node.MemoryCacheSize * 1024 * 1024)
+	cache := newCache(custom)
 	store, err := storage.NewBadgerStore(custom, dir)
 	assert.Nil(err)
 	assert.NotNil(store)
@@ -644,42 +477,108 @@ func testPledgeNewNode(assert *assert.Assertions, node string, domain common.Add
 	return Node{Signer: signer, Payee: payee, Host: "127.0.0.1:18099"}, pnode, server
 }
 
-func testBuildPledgeInput(assert *assert.Assertions, node string, domain common.Address, utxos []*common.VersionedTransaction) (string, error) {
-	inputs := []map[string]interface{}{}
+func testBuildPledgeInput(t *testing.T, nodes []*Node, domain common.Address, utxos []*common.VersionedTransaction, snapVersionMint int) (string, error) {
+	assert := assert.New(t)
+	inputs := []map[string]any{}
 	for _, tx := range utxos {
-		inputs = append(inputs, map[string]interface{}{
+		inputs = append(inputs, map[string]any{
 			"hash":  tx.PayloadHash().String(),
 			"index": 0,
 		})
 	}
-	raw, _ := json.Marshal(map[string]interface{}{
+	outputs := []map[string]any{}
+	for i := 0; i < NODES; i++ {
+		outputs = append(outputs, map[string]any{
+			"type":     0,
+			"amount":   common.NewIntegerFromString("3.5").Div(NODES),
+			"script":   "fffe01",
+			"accounts": []string{domain.String()},
+		})
+	}
+	outputs = append(outputs, map[string]any{
+		"type":     0,
+		"amount":   "10000",
+		"script":   "fffe01",
+		"accounts": []string{domain.String()},
+	})
+	raw, _ := json.Marshal(map[string]any{
 		"version": 2,
 		"asset":   "a99c2e0e2b1da4d648755ef19bd95139acbbe6564cfb06dec7cd34931ca72cdc",
 		"inputs":  inputs,
-		"outputs": []map[string]interface{}{{
-			"type":     0,
-			"amount":   "3.5",
-			"script":   "fffe01",
-			"accounts": []string{domain.String()},
-		}, {
-			"type":     0,
-			"amount":   "10000",
-			"script":   "fffe01",
-			"accounts": []string{domain.String()},
-		}},
+		"outputs": outputs,
 	})
-	tx, err := testSignTransaction(node, domain, string(raw))
+	tx, err := testSignTransaction(nodes[0].Host, domain, string(raw), snapVersionMint)
 	assert.Nil(err)
 	ver := common.VersionedTransaction{SignedTransaction: *tx}
-	input, err := testSendTransaction(node, hex.EncodeToString(ver.Marshal()))
-	assert.Nil(err)
-	var hash map[string]string
-	err = json.Unmarshal([]byte(input), &hash)
-	return hash["hash"], err
+	testSendTransactionsToNodesWithRetry(t, nodes, []*common.VersionedTransaction{&ver})
+	return ver.PayloadHash().String(), err
+}
+
+func testSendTransactionsToNodesWithRetry(t *testing.T, nodes []*Node, vers []*common.VersionedTransaction) {
+	assert := assert.New(t)
+
+	var wg sync.WaitGroup
+	for _, ver := range vers {
+		wg.Add(1)
+		go func(ver *common.VersionedTransaction) {
+			node := nodes[int(time.Now().UnixNano())%len(nodes)].Host
+			id, err := testSendTransaction(node, hex.EncodeToString(ver.Marshal()))
+			assert.Nil(err)
+			assert.Len(id, 75)
+			defer wg.Done()
+		}(ver)
+	}
+	wg.Wait()
+	time.Sleep(3 * time.Second)
+
+	var missingTxs []*common.VersionedTransaction
+	for _, ver := range vers {
+		node := nodes[int(time.Now().UnixNano())%len(nodes)].Host
+		data, _ := callRPC(node, "gettransaction", []any{ver.PayloadHash().String()})
+		var res map[string]string
+		json.Unmarshal([]byte(data), &res)
+		hash, _ := crypto.HashFromString(res["snapshot"])
+		if hash.HasValue() {
+			continue
+		}
+		t.Logf("TX MISSING %s\n", ver.PayloadHash())
+		missingTxs = append(missingTxs, ver)
+	}
+	if len(missingTxs) == 0 {
+		return
+	}
+	testSendTransactionsToNodesWithRetry(t, nodes, missingTxs)
+}
+
+func testSendTransactionToNodes(t *testing.T, nodes []*Node, raw string) string {
+	assert := assert.New(t)
+
+	rand.Seed(time.Now().UnixNano())
+	for n := len(nodes); n > 0; n-- {
+		ri := rand.Intn(n)
+		nodes[n-1], nodes[ri] = nodes[ri], nodes[n-1]
+	}
+	wg, dup := &sync.WaitGroup{}, 3
+	start := rand.Intn(len(nodes) - dup)
+
+	var txHash string
+	for i := start; i < len(nodes) && i != start+dup; i++ {
+		wg.Add(1)
+		go func(n string, raw string) {
+			defer wg.Done()
+			id, err := testSendTransaction(n, raw)
+			assert.Nil(err)
+			assert.Len(id, 75)
+			txHash = id
+		}(nodes[i].Host, raw)
+	}
+	wg.Wait()
+
+	return txHash
 }
 
 func testSendTransaction(node, raw string) (string, error) {
-	data, err := callRPC(node, "sendrawtransaction", []interface{}{
+	data, err := callRPC(node, "sendrawtransaction", []any{
 		raw,
 	})
 	return string(data), err
@@ -698,7 +597,7 @@ func testGetNodeToRemove(networkId crypto.Hash, signers, payees []common.Address
 	return nodes[seq][0], nodes[seq][1]
 }
 
-func testDeterminAccountByIndex(i int, role string) common.Address {
+func testDetermineAccountByIndex(i int, role string) common.Address {
 	seed := make([]byte, 64)
 	copy(seed, []byte("TESTNODE#"+role+"#"))
 	seed[63] = byte(i)
@@ -712,8 +611,8 @@ func setupTestNet(root string) ([]common.Address, []common.Address, []byte, stri
 	var signers, payees []common.Address
 
 	for i := 0; i < NODES; i++ {
-		signers = append(signers, testDeterminAccountByIndex(i, "SIGNER"))
-		payees = append(payees, testDeterminAccountByIndex(i, "PAYEE"))
+		signers = append(signers, testDetermineAccountByIndex(i, "SIGNER"))
+		payees = append(payees, testDetermineAccountByIndex(i, "PAYEE"))
 	}
 
 	inputs := make([]map[string]string, 0)
@@ -724,7 +623,7 @@ func setupTestNet(root string) ([]common.Address, []common.Address, []byte, stri
 			"balance": "10000",
 		})
 	}
-	genesis := map[string]interface{}{
+	genesis := map[string]any{
 		"epoch": 1551312000,
 		"nodes": inputs,
 		"domains": []map[string]string{
@@ -765,7 +664,7 @@ func setupTestNet(root string) ([]common.Address, []common.Address, []byte, stri
 	return signers, payees, genesisData, peersList
 }
 
-func testSignTransaction(node string, account common.Address, rawStr string) (*common.SignedTransaction, error) {
+func testSignTransaction(node string, account common.Address, rawStr string, snapVersionMint int) (*common.SignedTransaction, error) {
 	var raw signerInput
 	err := json.Unmarshal([]byte(rawStr), &raw)
 	if err != nil {
@@ -773,7 +672,10 @@ func testSignTransaction(node string, account common.Address, rawStr string) (*c
 	}
 	raw.Node = node
 
-	tx := common.NewTransaction(raw.Asset)
+	tx := common.NewTransactionV2(raw.Asset)
+	if snapVersionMint < 1 && time.Now().UnixNano()%2 == 1 {
+		tx = common.NewTransactionV3(raw.Asset)
+	}
 	for _, in := range raw.Inputs {
 		if d := in.Deposit; d != nil {
 			tx.AddDepositInput(&common.DepositData{
@@ -817,7 +719,6 @@ func testVerifyInfo(assert *assert.Assertions, nodes []*Node) Info {
 	info := testGetGraphInfo(nodes[0].Host)
 	for _, n := range nodes {
 		a := testGetGraphInfo(n.Host)
-		assert.Equal(info.Timestamp, a.Timestamp)
 		assert.Equal(info.PoolSize, a.PoolSize)
 	}
 	return info
@@ -834,13 +735,13 @@ func testVerifySnapshots(assert *assert.Assertions, nodes []*Node) (map[string]b
 		m, n := make(map[string]bool), make(map[string]bool)
 		for k := range a {
 			s[k] = true
-			t[a[k].Transaction.String()] = true
-			m[a[k].Transaction.String()] = true
+			t[a[k].SoleTransaction().String()] = true
+			m[a[k].SoleTransaction().String()] = true
 		}
 		for k := range b {
 			s[k] = true
-			t[b[k].Transaction.String()] = true
-			n[b[k].Transaction.String()] = true
+			t[b[k].SoleTransaction().String()] = true
+			n[b[k].SoleTransaction().String()] = true
 		}
 		assertKeyEqual(assert, a, b)
 		assert.Equal(len(a), len(b))
@@ -864,24 +765,49 @@ func assertKeyEqual(assert *assert.Assertions, a, b map[string]*common.Snapshot)
 }
 
 func testListSnapshots(node string) map[string]*common.Snapshot {
-	data, err := callRPC(node, "listsnapshots", []interface{}{
+	data, err := callRPC(node, "listsnapshots", []any{
 		0,
 		100000,
 		false,
 		false,
 	})
 
-	var snapshots []*common.Snapshot
-	if err != nil {
-		panic(err)
+	var rss []*struct {
+		Version           uint8                 `json:"version"`
+		NodeId            crypto.Hash           `json:"node_id"`
+		References        *common.RoundLink     `json:"references"`
+		RoundNumber       uint64                `json:"round_number"`
+		Timestamp         uint64                `json:"timestamp"`
+		Signatures        []*crypto.Signature   `json:"signatures"`
+		Signature         *crypto.CosiSignature `json:"signature"`
+		Hash              crypto.Hash           `json:"hash"`
+		Transactions      []crypto.Hash         `json:"transactions"`
+		TransactionLegacy crypto.Hash           `json:"transaction"`
 	}
-	err = json.Unmarshal(data, &snapshots)
+	err = json.Unmarshal(data, &rss)
 	if err != nil {
 		panic(err)
 	}
 	filter := make(map[string]*common.Snapshot)
-	for _, s := range snapshots {
-		filter[s.Hash.String()] = s
+	snapshots := make([]*common.Snapshot, len(rss))
+	for i, s := range rss {
+		snapshots[i] = &common.Snapshot{
+			Version:     s.Version,
+			NodeId:      s.NodeId,
+			RoundNumber: s.RoundNumber,
+			References:  s.References,
+			Timestamp:   s.Timestamp,
+		}
+		switch s.Version {
+		case 0:
+			snapshots[i].Signatures = s.Signatures
+		case 1, 2:
+			snapshots[i].Signature = s.Signature
+		default:
+			panic(s.Version)
+		}
+		snapshots[i].AddSoleTransaction(s.TransactionLegacy)
+		filter[s.Hash.String()] = snapshots[i]
 	}
 	return filter
 }
@@ -895,7 +821,7 @@ type Node struct {
 }
 
 func testListNodes(node string) []*Node {
-	data, err := callRPC(node, "listallnodes", []interface{}{0, false})
+	data, err := callRPC(node, "listallnodes", []any{0, false})
 	if err != nil {
 		panic(err)
 	}
@@ -914,7 +840,7 @@ type HeadRound struct {
 }
 
 func testDumpGraphHead(node string, id crypto.Hash) *HeadRound {
-	data, err := callRPC(node, "dumpgraphhead", []interface{}{})
+	data, err := callRPC(node, "dumpgraphhead", []any{})
 	if err != nil {
 		panic(err)
 	}
@@ -937,7 +863,7 @@ type Info struct {
 }
 
 func testGetGraphInfo(node string) Info {
-	data, err := callRPC(node, "getinfo", []interface{}{})
+	data, err := callRPC(node, "getinfo", []any{})
 	if err != nil {
 		panic(err)
 	}
@@ -963,12 +889,12 @@ func testGetGraphInfo(node string) Info {
 
 var httpClient *http.Client
 
-func callRPC(node, method string, params []interface{}) ([]byte, error) {
+func callRPC(node, method string, params []any) ([]byte, error) {
 	if httpClient == nil {
 		httpClient = &http.Client{Timeout: 30 * time.Second}
 	}
 
-	body, err := json.Marshal(map[string]interface{}{
+	body, err := json.Marshal(map[string]any{
 		"method": method,
 		"params": params,
 	})
@@ -989,8 +915,8 @@ func callRPC(node, method string, params []interface{}) ([]byte, error) {
 	defer resp.Body.Close()
 
 	var result struct {
-		Data  interface{} `json:"data"`
-		Error interface{} `json:"error"`
+		Data  any `json:"data"`
+		Error any `json:"error"`
 	}
 	err = json.NewDecoder(resp.Body).Decode(&result)
 	if err != nil {
@@ -1041,7 +967,7 @@ func (raw signerInput) ReadUTXOKeys(hash crypto.Hash, index int) (*common.UTXOKe
 		}
 	}
 
-	data, err := callRPC(raw.Node, "getutxo", []interface{}{hash.String(), index})
+	data, err := callRPC(raw.Node, "getutxo", []any{hash.String(), index})
 	if err != nil {
 		return nil, err
 	}
@@ -1065,3 +991,15 @@ func (raw signerInput) CheckDepositInput(deposit *common.DepositData, tx crypto.
 func (raw signerInput) ReadLastMintDistribution(group string) (*common.MintDistribution, error) {
 	return nil, nil
 }
+
+func newCache(conf *config.Custom) *ristretto.Cache {
+	cache, err := ristretto.NewCache(&ristretto.Config{
+		NumCounters: 1e7, // number of keys to track frequency of (10M).
+		MaxCost:     int64(conf.Node.MemoryCacheSize) * 1024 * 1024,
+		BufferItems: 64, // number of keys per Get buffer.
+	})
+	if err != nil {
+		panic(err)
+	}
+	return cache
+}
